"""
serialconfig_frame.py

Generic serial port configuration Frame subclass
for use in tkinter applications which require a
serial port configuration facility.

Exposes the serial port settings as properties.

Application icons from https://iconmonstr.com/license/.

Created on 24 Dec 2020

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

from tkinter import (
    DISABLED,
    HORIZONTAL,
    LEFT,
    NORMAL,
    VERTICAL,
    Button,
    Checkbutton,
    DoubleVar,
    E,
    Entry,
    Frame,
    IntVar,
    Label,
    Listbox,
    N,
    S,
    Scrollbar,
    Spinbox,
    StringVar,
    TclError,
    W,
)

from PIL import Image, ImageTk
from pyubx2 import GET, POLL, SET, SETPOLL
from serial import PARITY_EVEN, PARITY_MARK, PARITY_NONE, PARITY_ODD, PARITY_SPACE
from serial.tools.list_ports import comports

from pygpsclient.globals import (
    ICON_CONTRACT,
    ICON_EXPAND,
    ICON_REFRESH,
    LBAND,
    READONLY,
)
from pygpsclient.strings import LBLSPORT, LBLUDPORT

ADVOFF = "\u25bc"
ADVON = "\u25b2"
BGCOL = "azure"
BPSRATE_RNG = (9600, 19200, 38400, 57600, 115200, 230400, 460800, 921600, 1000000, 4800)
CONNECTED = 1
CONNECTED_FILE = 2
DATABITS_RNG = (8, 5, 6, 7)
DISCONNECTED = 0
MSGMODED = {
    GET: "GET",
    SET: "SET",
    POLL: "POLL",
    SETPOLL: "SETPOLL",
}
MSGMODE_RNG = list(MSGMODED.values())
NOPORTS = 3
PARITIES = {
    PARITY_NONE: "None",
    PARITY_EVEN: "Even",
    PARITY_ODD: "Odd",
    PARITY_MARK: "Mark",
    PARITY_SPACE: "Space",
}
PARITY_RNG = list(PARITIES.values())
STOPBITS_RNG = (1, 1.5, 2)
TIMEOUT_RNG = ("0", "1", "2", "5", "10", "20")


class SerialConfigFrame(Frame):
    """
    Serial port configuration frame class.
    """

    def __init__(self, app, container, context, *args, **kwargs):
        """
        Constructor.

        :param Frame app: reference to main tkinter application
        :param tkinter.Frame container: reference to container frame
        :param str context: serial port context (GNSS or LBAND)
        :param args: optional args to pass to Frame parent class
        :param kwargs: optional kwargs for value ranges, or to pass to Frame parent class
        """

        self._bpsrate_rng = kwargs.pop("bpsrates", BPSRATE_RNG)
        self._databits_rng = kwargs.pop("databits", DATABITS_RNG)
        self._stopbits_rng = kwargs.pop("stopbits", STOPBITS_RNG)
        self._parity_rng = kwargs.pop("parities", PARITY_RNG)
        self._timeout_rng = kwargs.pop("timeouts", TIMEOUT_RNG)
        self._msgmode_name_rng = kwargs.pop("msgmodes", MSGMODE_RNG)
        self._preselect = kwargs.pop("preselect", ())
        self._context = context

        Frame.__init__(self, container, *args, **kwargs)

        self.__app = app
        self._show_advanced = False
        self._status = DISCONNECTED
        self._ports = ()
        self._port = StringVar()
        self.user_defined_port = StringVar()
        self.user_defined_port.set(self.__app.configuration.get("userport_s"))
        self._port_desc = StringVar()
        self._bpsrate = IntVar()
        self._databits = IntVar()
        self._stopbits = DoubleVar()
        self._parity_name = StringVar()
        self._rtscts = IntVar()
        self._xonxoff = IntVar()
        self._timeout = DoubleVar()
        self._msgmode_name = StringVar()
        self._inactivity_timeout = IntVar()
        self._img_refresh = ImageTk.PhotoImage(Image.open(ICON_REFRESH))
        self._img_expand = ImageTk.PhotoImage(Image.open(ICON_EXPAND))
        self._img_contract = ImageTk.PhotoImage(Image.open(ICON_CONTRACT))

        self._body()
        self._do_layout()
        self._attach_events()
        self.reset()

    def _body(self):
        """
        Set up widgets.
        """

        self._frm_basic = Frame(self)
        self._lbl_port = Label(self._frm_basic, text="Serial\nPort  ")
        self._lbx_port = Listbox(
            self._frm_basic,
            border=2,
            relief="sunken",
            width=38,
            height=5,
            justify=LEFT,
            exportselection=False,
        )
        self._scr_portv = Scrollbar(self._frm_basic, orient=VERTICAL)
        self._scr_porth = Scrollbar(self._frm_basic, orient=HORIZONTAL)
        self._lbx_port.config(yscrollcommand=self._scr_portv.set)
        self._lbx_port.config(xscrollcommand=self._scr_porth.set)
        self._scr_portv.config(command=self._lbx_port.yview)
        self._scr_porth.config(command=self._lbx_port.xview)
        self._lbl_bpsrate = Label(self._frm_basic, text="Rate bps")
        self._spn_bpsrate = Spinbox(
            self._frm_basic,
            values=self._bpsrate_rng,
            width=8,
            state=READONLY,
            wrap=True,
            textvariable=self._bpsrate,
        )
        self._btn_refresh = Button(
            self._frm_basic,
            command=self._on_refresh,
            image=self._img_refresh,
            width=28,
            height=22,
        )
        self._btn_toggle = Button(
            self._frm_basic,
            command=self._on_toggle_advanced,
            image=self._img_expand,
            width=28,
            height=22,
        )

        self._frm_advanced = Frame(self)
        self._lbl_databits = Label(self._frm_advanced, text="Data Bits")
        self._spn_databits = Spinbox(
            self._frm_advanced,
            values=self._databits_rng,
            width=3,
            state=READONLY,
            wrap=True,
            textvariable=self._databits,
        )
        self._lbl_stopbits = Label(self._frm_advanced, text="Stop Bits")
        self._spn_stopbits = Spinbox(
            self._frm_advanced,
            values=self._stopbits_rng,
            width=3,
            state=READONLY,
            wrap=True,
            textvariable=self._stopbits,
        )
        self._lbl_parity_name = Label(self._frm_advanced, text="Parity")
        self._spn_parity_name = Spinbox(
            self._frm_advanced,
            values=self._parity_rng,
            width=6,
            state=READONLY,
            wrap=True,
            textvariable=self._parity_name,
        )
        self._chk_rts = Checkbutton(
            self._frm_advanced, text="RTS/CTS", variable=self._rtscts
        )
        self._chk_xon = Checkbutton(
            self._frm_advanced, text="Xon/Xoff", variable=self._xonxoff
        )
        self._lbl_timeout = Label(self._frm_advanced, text="Timeout (s)")
        self._spn_timeout = Spinbox(
            self._frm_advanced,
            values=self._timeout_rng,
            width=4,
            state=READONLY,
            wrap=True,
            textvariable=self._timeout,
        )
        self._lbl_msgmode_name = Label(self._frm_advanced, text="Msg Mode")
        self._spn_msgmode_name = Spinbox(
            self._frm_advanced,
            values=self._msgmode_name_rng,
            width=7,
            state=READONLY,
            wrap=True,
            textvariable=self._msgmode_name,
        )
        self._lbl_userport = Label(
            self._frm_advanced, text="User-defined\nPort               "
        )
        self._ent_userport = Entry(
            self._frm_advanced,
            textvariable=self.user_defined_port,
            relief="sunken",
            width=30,
        )
        self._lbl_inactivity = Label(self._frm_advanced, text="Inactivity Timeout (s)")
        self._spn_inactivity = Spinbox(
            self._frm_advanced,
            values=(0, 1, 5, 10, 30, 60, 300),
            width=4,
            state=READONLY,
            wrap=True,
            textvariable=self._inactivity_timeout,
        )

    def _do_layout(self):
        """
        Layout widgets.
        """

        self._frm_basic.grid(column=0, row=0, columnspan=4, sticky=(W, E))
        self._lbl_port.grid(column=0, row=0, sticky=W)
        self._lbx_port.grid(
            column=1, row=0, columnspan=3, sticky=(W, E), padx=3, pady=2
        )
        self._scr_portv.grid(column=4, row=0, sticky=(N, S))
        self._scr_porth.grid(column=1, row=1, columnspan=3, sticky=(E, W))
        self._lbl_bpsrate.grid(column=0, row=2, sticky=W)
        self._spn_bpsrate.grid(column=1, row=2, sticky=W, padx=3, pady=2)
        self._btn_refresh.grid(column=3, row=2, sticky=E)
        self._btn_toggle.grid(column=4, row=2, sticky=E)

        self._frm_advanced.grid_forget()
        self._lbl_databits.grid(column=0, row=0, sticky=W)
        self._spn_databits.grid(column=1, row=0, sticky=W, padx=3, pady=2)
        self._lbl_stopbits.grid(column=2, row=0, sticky=W)
        self._spn_stopbits.grid(column=3, row=0, sticky=W, padx=3, pady=2)
        self._lbl_parity_name.grid(column=0, row=1, sticky=W)
        self._spn_parity_name.grid(column=1, row=1, sticky=W, padx=3, pady=2)
        self._chk_rts.grid(column=2, row=1, sticky=W)
        self._chk_xon.grid(column=3, row=1, sticky=W, padx=3, pady=2)
        self._lbl_timeout.grid(column=0, row=2, sticky=W)
        self._spn_timeout.grid(column=1, row=2, sticky=W, padx=3, pady=2)
        if len(self._msgmode_name_rng) > 0:
            self._lbl_msgmode_name.grid(column=2, row=2, sticky=W)
            self._spn_msgmode_name.grid(column=3, row=2, sticky=W, padx=3, pady=2)
        self._lbl_userport.grid(column=0, row=3, sticky=W)
        self._ent_userport.grid(column=1, row=3, columnspan=3, sticky=W, padx=3, pady=2)
        self._lbl_inactivity.grid(column=0, row=4, columnspan=2, sticky=W)
        self._spn_inactivity.grid(
            column=2, row=4, columnspan=3, sticky=W, padx=3, pady=2
        )

    def _attach_events(self):
        """
        Bind events to frame.
        """

        self.bind("<Configure>", self._on_resize)
        self._lbx_port.bind("<<ListboxSelect>>", self._on_select_port)

    def _bind_events(self, add: bool = True):
        """
        Add or remove event bindings to/from widgets.

        :param bool add: add or remove binding
        """

        tracemode = "write"
        for setting in (
            self._port,
            self.user_defined_port,
            self._bpsrate,
            self._databits,
            self._stopbits,
            self._parity_name,
            self._rtscts,
            self._xonxoff,
            self._timeout,
            self._msgmode_name,
            self._inactivity_timeout,
        ):
            if add:
                setting.trace_add(tracemode, self._on_update_config)
            else:
                if len(setting.trace_info()) > 0:
                    setting.trace_remove(tracemode, setting.trace_info()[0][1])

    def reset(self):
        """
        Reset settings to saved configuration.
        """

        self._bind_events(False)
        cfg = self.__app.configuration
        if self._context == LBAND:
            self._port.set(cfg.get("lbandclientserialport_s"))
            self._bpsrate.set(cfg.get("lbandclientbpsrate_n"))
            self._databits.set(cfg.get("lbandclientdatabits_n"))
            self._stopbits.set(cfg.get("lbandclientstopbits_f"))
            self._parity_name.set(PARITIES[cfg.get("lbandclientparity_s")])
            self._rtscts.set(cfg.get("lbandclientrtscts_b"))
            self._xonxoff.set(cfg.get("lbandclientxonxoff_b"))
            self._timeout.set(cfg.get("lbandclienttimeout_f"))
            self._msgmode_name.set(MSGMODED[cfg.get("lbandclientmsgmode_n")])
            self._inactivity_timeout.set(cfg.get("lbandclientinactivity_timeout_n"))
            self.user_defined_port.set(cfg.get("spartnport_s"))
        else:  # GNSS
            self._port.set(cfg.get("serialport_s"))
            self._bpsrate.set(cfg.get("bpsrate_n"))
            self._databits.set(cfg.get("databits_n"))
            self._stopbits.set(cfg.get("stopbits_f"))
            self._parity_name.set(PARITIES[cfg.get("parity_s")])
            self._rtscts.set(cfg.get("rtscts_b"))
            self._xonxoff.set(cfg.get("xonxoff_b"))
            self._timeout.set(cfg.get("timeout_f"))
            self._msgmode_name.set(MSGMODED[cfg.get("msgmode_n")])
            self._inactivity_timeout.set(cfg.get("inactivity_timeout_n"))
            self.user_defined_port.set(cfg.get("userport_s"))
        self._on_refresh()
        self._bind_events(True)

    def _on_update_config(self, var, index, mode):  # pylint: disable=unused-argument
        """
        Update in-memory configuration if setting is changed.
        """

        try:
            self.update()
            cfg = self.__app.configuration
            if self._context == LBAND:
                cfg.set("lbandclientserialport_s", self.port)
                cfg.set("spartnport_s", self.userport)
                cfg.set("lbandclientbpsrate_n", int(self.bpsrate))
                cfg.set("lbandclientdatabits_n", int(self.databits))
                cfg.set("lbandclientstopbits_f", float(self.stopbits))
                cfg.set("lbandclientparity_s", self.parity)
                cfg.set("lbandclientrtscts_b", int(self.rtscts))
                cfg.set("lbandclientxonxoff_b", int(self.xonxoff))
                cfg.set("lbandclienttimeout_f", float(self.timeout))
                cfg.set("lbandclientmsgmode_n", int(self.msgmode))
                cfg.set("lbandclientinactivity_timeout_n", int(self.inactivity_timeout))
            else:  # GNSS
                cfg.set("serialport_s", self.port)
                cfg.set("userport_s", self.userport)
                cfg.set("bpsrate_n", int(self.bpsrate))
                cfg.set("databits_n", int(self.databits))
                cfg.set("stopbits_f", float(self.stopbits))
                cfg.set("parity_s", self.parity)
                cfg.set("rtscts_b", int(self.rtscts))
                cfg.set("xonxoff_b", int(self.xonxoff))
                cfg.set("timeout_f", float(self.timeout))
                cfg.set("msgmode_n", int(self.msgmode))
                cfg.set("inactivity_timeout_n", int(self.inactivity_timeout))
        except (ValueError, TclError):
            pass

    def _on_refresh(self):
        """
        Refresh list of ports.
        """

        if self._status in (CONNECTED, CONNECTED_FILE):
            return
        self.set_status(DISCONNECTED)
        self._lbx_port.delete(0, "end")
        self._get_ports()

    def _get_ports(self):
        """
        Populate list of available serial ports using pyserial comports tool.

        User-defined serial port can be passed as command line keyword argument,
        in which case this takes precedence.

        Attempt to automatically select a serial device matching
        a list of 'preselect' devices (only works on platforms
        which parse UART device desc or HWID e.g. Posix).
        """

        self._ports = sorted(comports())
        init_idx = 0
        recognised = False
        pnames = [p[0] for p in self._ports]
        userp = self.user_defined_port.get()
        if self._port.get() != "" and self._port.get() not in pnames:
            self._ports.insert(0, (self._port.get(), LBLSPORT, None))
        if userp != "" and userp not in pnames:
            self._ports.insert(0, (userp, LBLUDPORT, None))

        if len(self._ports) > 0:
            # default to first item in list
            port, desc, _ = self._ports[0]
            if desc == "":
                desc = "device"
            self._port.set(port)
            self._port_desc.set(desc)
            for idx, (port, desc, _) in enumerate(self._ports):
                self._lbx_port.insert(idx, f"{port}: {desc}")
                # default selection to recognised GNSS device if possible
                if not recognised:
                    for dev in self._preselect:
                        if dev.lower() in desc.lower():
                            init_idx = idx
                            self._port.set(port)
                            self._port_desc.set(desc)
                            recognised = True
                            break
            self.set_status(DISCONNECTED)
            self._lbx_port.activate(init_idx)
            self._lbx_port.selection_set(first=init_idx)
        else:
            self.set_status(NOPORTS)

    def _on_select_port(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Get selected port from listbox and set global variable.
        """

        idx = self._lbx_port.curselection()
        if idx == "":
            idx = 0
        port_orig = self._lbx_port.get(idx)
        port = port_orig[0 : port_orig.find(":")]
        desc = port_orig[port_orig.find(":") + 1 :]
        if desc == "":
            desc = "device"
        self._port.set(port)
        self._port_desc.set(desc)

    def _on_toggle_advanced(self):
        """
        Toggle advanced serial port settings panel on or off.
        """

        self._show_advanced = not self._show_advanced
        if self._show_advanced:
            self._frm_advanced.grid(column=0, row=1, columnspan=3, sticky=(W, E))
            self._btn_toggle.config(image=self._img_contract)
        else:
            self._frm_advanced.grid_forget()
            self._btn_toggle.config(image=self._img_expand)

    def set_status(self, status: int = DISCONNECTED):
        """
        Set connection status, which determines whether controls
        are enabled or not: 0=DISCONNECTED, 1=CONNECTED,
        2=CONNECTED_FILE, 3=NOPORTS.

        :param int status: status (0,1,2,3)
        """

        self._status = status
        for widget in (
            self._lbl_port,
            self._lbl_bpsrate,
            self._lbl_databits,
            self._lbl_stopbits,
            self._lbl_parity_name,
            self._lbl_timeout,
            self._chk_rts,
            self._chk_xon,
            self._lbx_port,
            self._lbl_msgmode_name,
            self._lbl_userport,
            self._ent_userport,
            self._lbl_inactivity,
        ):
            widget.configure(state=(NORMAL if status == DISCONNECTED else DISABLED))
        for widget in (
            self._spn_bpsrate,
            self._spn_databits,
            self._spn_stopbits,
            self._spn_parity_name,
            self._spn_timeout,
            self._spn_msgmode_name,
            self._spn_inactivity,
        ):
            widget.configure(state=(READONLY if status == DISCONNECTED else DISABLED))

    @property
    def status(self) -> int:
        """
        Getter for status flag: 0=DISCONNECTED, 1=CONNECTED,
        2=CONNECTED_FILE, 3=NOPORTS.

        :return: status flag (0,1,2,3)
        :rtype: int
        """

        return self._status

    @property
    def port(self) -> str:
        """
        Getter for port.

        :return: selected port
        :rtype: str
        """

        return self._port.get()

    @property
    def port_desc(self) -> str:
        """
        Getter for port description.

        :return: selected port description
        :rtype: str
        """

        return self._port_desc.get()

    @property
    def bpsrate(self) -> int:
        """
        Getter for bps rate (commonly but incorrectly referred to as baud rate).

        :return: selected baudrate
        :rtype: int
        """

        return self._bpsrate.get()

    @property
    def databits(self) -> int:
        """
        Getter for databits.

        :return: selected databits
        :rtype: int
        """

        return self._databits.get()

    @property
    def stopbits(self) -> float:
        """
        Getter for stopbits.

        :return: selected stopbits
        :rtype: float
        """

        return self._stopbits.get()

    @property
    def parity(self) -> str:
        """
        Getter for parity.

        :return: selected parity
        :rtype: str
        """

        if self._parity_name.get() == PARITIES[PARITY_EVEN]:
            return PARITY_EVEN
        if self._parity_name.get() == PARITIES[PARITY_ODD]:
            return PARITY_ODD
        if self._parity_name.get() == PARITIES[PARITY_MARK]:
            return PARITY_MARK
        if self._parity_name.get() == PARITIES[PARITY_SPACE]:
            return PARITY_SPACE
        return PARITY_NONE

    @property
    def rtscts(self) -> bool:
        """
        Getter for rts/cts.

        :return: selected rts/cts
        :rtype: bool
        """

        return self._rtscts.get()

    @property
    def xonxoff(self) -> bool:
        """
        Getter for xon/xoff.

        :return: selected xon/xoff
        :rtype: bool
        """

        return self._xonxoff.get()

    @property
    def timeout(self) -> float:
        """
        Getter for timeout.

        :return: selected timeout
        :rtype: float (or None)
        """

        if self._timeout.get() == "None":
            return None
        return float(self._timeout.get())

    @property
    def msgmode(self) -> int:
        """
        Return message parsing mode
        Default is GET i.e. input from receiver.

        :return: message mode 0 = POLL 1 = SET, 2 = POLL, 3 = SETPOLL
        :rtype: int
        """

        if self._msgmode_name.get() == MSGMODED[SET]:
            return SET
        if self._msgmode_name.get() == MSGMODED[POLL]:
            return POLL
        if self._msgmode_name.get() == MSGMODED[SETPOLL]:
            return SETPOLL
        return GET

    @property
    def userport(self) -> str:
        """
        Return user-defined port

        :return: user-defined serial port
        :rtype: str
        """

        return self.user_defined_port.get()

    @property
    def inactivity_timeout(self) -> int:
        """
        Return inactivity timeout

        :return: timeout in seconds
        :rtype: int
        """

        return self._inactivity_timeout.get()

    def _on_resize(self, event):  # pylint: disable=unused-argument
        """
        Resize frame.

        :param event event: resize event
        """

        self.__app.frm_settings.on_expand()



"""
chart_frame.py

Chart frame class for PyGPSClient application.

This emulates a 4-channel "oscilloscope", allowing the user to plot
up to 4 named data attributes over time. X-axis and Y-axis scale and
ranges are all configurable.

Plot limited only by the number of data points that can be comfortably
held in memory.

Created on 24 Nov 2024

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

from random import choice
from time import time
from tkinter import (
    ALL,
    NE,
    NORMAL,
    NW,
    SE,
    SW,
    Canvas,
    E,
    Entry,
    Frame,
    Label,
    N,
    S,
    Spinbox,
    StringVar,
    TclError,
    W,
)

from pygpsclient.globals import (
    AXISCOL,
    BGCOL,
    GRIDCOL,
    READONLY,
    WIDGETU6,
    AreaXY,
)
from pygpsclient.helpers import data2xy, fontheight, get_grid, scale_font, time2str

MAXCHANS = 4
RESFONT = 28  # font size relative to widget size
MINFONT = 8  # minimum font size
PLOTWID = 1
PLOTCOLS = ("yellow", "cyan", "magenta", "deepskyblue")
GRIDMINCOL = "grey30"
LBLGRID = 5
GRIDSTEPS = get_grid(21)
XLBLSTEPS = get_grid(LBLGRID)
YLBLSTEPS = get_grid(LBLGRID)
ERRCOL = "coral"
LBLCOL = "white"
CONTRASTCOL = "black"
MODEINIT = "chart"
AXISTAG = "axt"
AXISLBLTAG = "axl"
# total capacity depends on available free memory...
TIMRANGE = [int(i * 10**n) for n in (1, 2, 3, 4) for i in (1, 2.4, 3.6, 4.8, 6)]
DPTRANGE = [int(i * 10**n) for n in (3, 4, 5, 6) for i in (1, 2, 5)]
XRANGE = 120
CHARTMINY = 0
CHARTMAXY = 100
CHARTSCALE = 1
MINY = "MinY {}"
MAXY = "MaxY {}"


def gen_yrange() -> tuple:
    """
    Generate scale and max/min Y ranges for spinboxes.

    :return: Y range
    :rtype: tuple
    """

    srange = ()
    for i in range(0, 8):
        for n in (1, 2, 5):
            srange += (n * 10**i,)
    for i in range(8, 0, -1):
        for n in (5, 2, 1):
            v = n * 10**-i
            if v == 4.9999999999999996e-06:  # fix Python rounding quirk!
                v = 5e-06
            srange += (v,)

    yrange = ("0",)
    for i in range(0, 8):
        for n in (1, 2, 5):
            yrange += (n * 10**i,)
    for i in range(8, -1, -1):
        for n in (5, 2, 1):
            yrange += (-n * 10**i,)

    return srange, yrange


class ChartviewFrame(Frame):
    """
    CHartview frame class.
    """

    def __init__(self, app, *args, **kwargs):
        """
        Constructor.

        :param Frame app: reference to main tkinter application
        :param args: optional args to pass to Frame parent class
        :param kwargs: optional kwargs to pass to Frame parent class
        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.appmaster  # Reference to root class (Tk)

        Frame.__init__(self, self.__master, *args, **kwargs)

        self.chartsettings = self.__app.configuration.get("chartsettings_d")
        def_w, def_h = WIDGETU6
        self.width = kwargs.get("width", def_w)
        self.height = kwargs.get("height", def_h)
        self.configure(bg=BGCOL)
        self._font = None
        self._fonth = 6
        self._xoff = 20  # chart X offset for labels
        self._yoff = 20  # chart Y offset for labels
        self._chart_data = {}
        self._num_chans = self.chartsettings.get("numchn_n", MAXCHANS)
        if self._num_chans % 2:  # no channels must be even
            self._num_chans += 1
        self._plotcols = PLOTCOLS
        self._font = self.__app.font_sm
        self._fonth = fontheight(self._font)
        # generate random plot colours for channels > 4
        if self._num_chans > 4:
            self._plotcols += tuple(
                "#" + "".join([choice("9ABCDEF") for j in range(6)])
                for i in range(self._num_chans - 4)
            )
        self._data_id = [None] * self._num_chans
        self._data_name = [None] * self._num_chans
        self._data_scale = [None] * self._num_chans
        self._data_miny = [None] * self._num_chans
        self._data_maxy = [None] * self._num_chans
        self._mintim = 1e20
        self._maxtim = 0
        self._timrange = StringVar()
        self._maxpoints = StringVar()
        for chn in range(self._num_chans):
            self._data_id[chn] = StringVar()
            self._data_name[chn] = StringVar()
            self._data_scale[chn] = StringVar()
            self._data_miny[chn] = StringVar()
            self._data_maxy[chn] = StringVar()
        self._body()
        self._do_layout()
        self.reset()
        self._attach_events()

    def _body(self):
        """
        Set up frame and widgets.
        """

        srange, yrange = gen_yrange()
        # set column and row expand behaviour
        for i in range(6):
            self.grid_columnconfigure(i, weight=1, uniform="ent")
        self.grid_rowconfigure(0, weight=1)
        for i in range(1, 2 + self._num_chans):
            self.grid_rowconfigure(i, weight=0)
        self._can_chartview = Canvas(
            self, width=self.width, height=self.height, bg=BGCOL
        )
        self._lbl_id = Label(
            self,
            text="Identity",
            fg=LBLCOL,
            bg=BGCOL,
        )
        self._lbl_name = Label(
            self,
            text="Name",
            fg=LBLCOL,
            bg=BGCOL,
        )
        self._lbl_scale = Label(
            self,
            text="Scale",
            fg=LBLCOL,
            bg=BGCOL,
        )
        self._lbl_miny = Label(
            self,
            text=MINY.format(""),
            fg=LBLCOL,
            bg=BGCOL,
        )
        self._lbl_maxy = Label(
            self,
            text=MAXY.format(""),
            fg=LBLCOL,
            bg=BGCOL,
        )
        self._lbl_timrange = Label(
            self,
            text="Time Range s",
            fg=LBLCOL,
            bg=BGCOL,
        )
        self._lbl_maxpoints = Label(
            self,
            text="Max Points",
            fg=LBLCOL,
            bg=BGCOL,
        )
        self._ent_id = [None] * self._num_chans
        self._ent_name = [None] * self._num_chans
        self._spn_scale = [None] * self._num_chans
        self._spn_miny = [None] * self._num_chans
        self._spn_maxy = [None] * self._num_chans
        for chn in range(self._num_chans):
            self._ent_id[chn] = Entry(
                self,
                textvariable=self._data_id[chn],
                state=NORMAL,
                relief="sunken",
                width=10,
                fg=self._plotcols[chn],
                bg=BGCOL,
            )
            self._ent_name[chn] = Entry(
                self,
                textvariable=self._data_name[chn],
                state=NORMAL,
                relief="sunken",
                width=10,
                fg=self._plotcols[chn],
                bg=BGCOL,
            )
            self._spn_scale[chn] = Spinbox(
                self,
                values=srange,
                wrap=True,
                textvariable=self._data_scale[chn],
                state=NORMAL,
                width=10,
                fg=self._plotcols[chn],
                bg=BGCOL,
                buttonbackground=BGCOL,
            )
            self._spn_miny[chn] = Spinbox(
                self,
                values=yrange,
                wrap=True,
                textvariable=self._data_miny[chn],
                state=NORMAL,
                width=10,
                fg=self._plotcols[chn],
                bg=BGCOL,
                buttonbackground=BGCOL,
            )
            self._spn_maxy[chn] = Spinbox(
                self,
                values=yrange,
                wrap=True,
                textvariable=self._data_maxy[chn],
                state=NORMAL,
                width=10,
                fg=self._plotcols[chn],
                bg=BGCOL,
                buttonbackground=BGCOL,
            )

        self._spn_timrange = Spinbox(
            self,
            values=TIMRANGE,
            wrap=True,
            textvariable=self._timrange,
            state=READONLY,
            width=8,
            fg=LBLCOL,
            readonlybackground=BGCOL,
            buttonbackground=BGCOL,
        )
        self._spn_maxpoints = Spinbox(
            self,
            values=DPTRANGE,
            wrap=True,
            textvariable=self._maxpoints,
            state=READONLY,
            width=8,
            fg=LBLCOL,
            readonlybackground=BGCOL,
            buttonbackground=BGCOL,
        )

    def _do_layout(self):
        """
        Position widgets in frame.
        """

        self._can_chartview.grid(column=0, row=0, columnspan=6, sticky=(N, S, E, W))
        self._lbl_id.grid(column=0, row=1, sticky=(W, E))
        self._lbl_name.grid(column=1, row=1, sticky=(W, E))
        self._lbl_scale.grid(column=2, row=1, sticky=(W, E))
        self._lbl_miny.grid(column=3, row=1, sticky=(W, E))
        self._lbl_maxy.grid(column=4, row=1, sticky=(W, E))
        for chn in range(self._num_chans):
            self._ent_id[chn].grid(column=0, row=2 + chn, sticky=(W, E))
            self._ent_name[chn].grid(column=1, row=2 + chn, sticky=(W, E))
            self._spn_scale[chn].grid(column=2, row=2 + chn, sticky=(W, E))
            self._spn_miny[chn].grid(column=3, row=2 + chn, sticky=(W, E))
            self._spn_maxy[chn].grid(column=4, row=2 + chn, sticky=(W, E))
        self._lbl_timrange.grid(column=5, row=1, sticky=(W, E))
        self._spn_timrange.grid(column=5, row=2, sticky=(W, E))
        self._lbl_maxpoints.grid(column=5, row=3, sticky=(W, E))
        self._spn_maxpoints.grid(column=5, row=4, sticky=(W, E))

    def _attach_events(self):
        """
        Bind events to frame.
        """

        self.bind("<Configure>", self._on_resize)
        self._can_chartview.bind("<Double-Button-1>", self._on_clear)
        self._can_chartview.bind("<Double-Button-2>", self._on_clipboard)
        self._can_chartview.bind("<Double-Button-3>", self._on_clipboard)
        self._timrange.trace_add("write", self._on_update_config)
        self._maxpoints.trace_add("write", self._on_update_config)
        for chn in range(self._num_chans):
            self._data_id[chn].trace_add("write", self._on_update_config)
            self._data_name[chn].trace_add("write", self._on_update_config)
            self._data_scale[chn].trace_add("write", self._on_update_config)
            self._data_miny[chn].trace_add("write", self._on_update_config)
            self._data_maxy[chn].trace_add("write", self._on_update_config)

    def _on_update_config(self, var, index, mode):  # pylint: disable=unused-argument
        """
        Update in-memory configuration if setting is changed.
        """

        try:
            cst = {}
            cst["numchn_n"] = int(self._num_chans)
            cst["timrng_n"] = int(self._timrange.get())
            cst["maxpoints_n"] = int(self._maxpoints.get())
            for chn in range(self._num_chans):
                cst[chn] = {}
                cst[chn]["id_s"] = self._data_id[chn].get()
                cst[chn]["name_s"] = self._data_name[chn].get()
                cst[chn]["scale_f"] = float(self._data_scale[chn].get())
                cst[chn]["miny_f"] = float(self._data_miny[chn].get())
                cst[chn]["maxy_f"] = float(self._data_maxy[chn].get())
            self.__app.configuration.set("chartsettings_d", cst)
        except (ValueError, TclError):
            pass

    def reset(self):
        """
        Reset chart frame.
        """

        self._timrange.set(self.chartsettings.get("timrng_n", TIMRANGE[3]))  # 60s
        self._maxpoints.set(self.chartsettings.get("maxpoints_n", DPTRANGE[2]))  # 5000
        for chn in range(self._num_chans):
            cst = self.chartsettings.get(str(chn), {})
            self._data_id[chn].set(cst.get("id_s", ""))
            self._data_name[chn].set(cst.get("name_s", ""))
            self._data_scale[chn].set(cst.get("scale_f", 1))
            self._data_miny[chn].set(cst.get("miny_f", CHARTMINY))
            self._data_maxy[chn].set(cst.get("maxy_f", CHARTMAXY))

        self._on_clear(None)

    def _on_clear(self, event):  # pylint: disable=unused-argument
        """
        Clear data.
        """

        self._chart_data = {}
        self._mintim = 1e20
        self._maxtim = 0
        self._can_chartview.delete(ALL)
        self.update_frame()

    def _valid_settings(self) -> bool:
        """
        Validate settings.

        :return: True/False
        :rtype: bool
        """

        try:
            for chn in range(self._num_chans):
                _ = float(self._data_scale[chn].get())
                _ = float(self._data_miny[chn].get())
                _ = float(self._data_maxy[chn].get())
            _ = float(self._timrange.get())
            return True
        except ValueError:
            self._disp_alert("Error: Invalid settings")
            return False

    def init_frame(self):
        """
        Initialise spectrum chart.
        """

        w, h = self.width, self.height
        self._xoff = self._fonth * self._num_chans / 2 + 3  # chart X offset for labels
        self._yoff = self._fonth + 3  # chart Y offset for labels
        self._can_chartview.delete(ALL)

        # draw grid
        for i, p in enumerate(GRIDSTEPS):
            y = (h - self._yoff) * p
            col = AXISCOL if p in (0, 1.0) else GRIDMINCOL if i % LBLGRID else GRIDCOL
            self._can_chartview.create_line(
                self._xoff, y, w - self._xoff, y, fill=col, tags=AXISTAG
            )
            x = self._xoff + (w - self._xoff * 2) * p
            self._can_chartview.create_line(
                x, 0, x, h - self._yoff, fill=col, tags=AXISTAG
            )

    def update_data(self, parsed_data: object):
        """
        Update chart data from parsed message.

        :param object parsed_data: parsed message
        """

        try:
            maxpoints = int(self._maxpoints.get())
        except ValueError:
            maxpoints = DPTRANGE[2]  # 5000

        now = round(time(), 0)  # time to nearest second
        if now not in self._chart_data:
            self._chart_data[now] = {}
        for chn in range(self._num_chans):
            mid = self._data_id[chn].get()
            name = self._data_name[chn].get()
            if name == "":
                continue
            if mid != "":
                if hasattr(parsed_data, "identity"):
                    if parsed_data.identity != mid:
                        continue

            # wildcards *+-, sum, max or min of group of values
            if name[-1] in ("*", "+", "-"):
                vals = []
                for attr in parsed_data.__dict__:
                    if name[:-1] in attr and name[0] != "_":
                        try:
                            vals.append(float(getattr(parsed_data, attr)))
                        except ValueError:
                            continue
                if vals:  # != []
                    if name[-1] == "+":
                        val = max(vals)
                    elif name[-1] == "-":
                        val = min(vals)
                    else:
                        val = sum(vals) / (len(vals) * 1.0)
                else:
                    val = None
            else:
                if hasattr(parsed_data, name):
                    try:
                        val = float(getattr(parsed_data, name))
                    except ValueError:
                        val = None
                else:
                    continue

            self._chart_data[now][chn] = val

            # update X axis (time) range
            self._mintim = min(now, self._mintim)
            self._maxtim = max(now, self._maxtim)

            # flag if scaled value is out of range
            self.flag_outofrange(chn, val)

        # limit number of data points
        while len(self._chart_data) > maxpoints:
            self._chart_data.pop(min(self._chart_data))

    def flag_outofrange(self, chn: int, val: float):
        """
        Flag if scaled value is over or under range.

        :param int chn: channel number
        :param float val: value
        """

        if val is None:
            return

        try:
            minval = float(self._data_miny[chn].get())
            maxval = float(self._data_maxy[chn].get())
            scale = float(self._data_scale[chn].get())
            ucol = ocol = BGCOL
            ufcol = ofcol = self._plotcols[chn]
            if val / scale < minval:
                ucol = ERRCOL
                ufcol = CONTRASTCOL
            elif val / scale > maxval:
                ocol = ERRCOL
                ofcol = CONTRASTCOL
            self._spn_miny[chn].configure(bg=ucol, fg=ufcol)
            self._spn_maxy[chn].configure(bg=ocol, fg=ofcol)
        except (TypeError, ValueError):
            pass

    def update_frame(self):
        """
        Plot selected chart data.
        """

        self._update_plot(self._chart_data)

    def _update_plot(self, data: dict):
        """
        Update chart plot with data.

        :param dict data: list of chart data
        :param int xrange: number of points
        """

        if not self._valid_settings():
            return

        mintim = 1e20
        maxtim = 0
        w, h = self.width, self.height
        self.init_frame()

        # set default ranges for all channels
        minval = [CHARTMINY] * self._num_chans
        maxval = [CHARTMAXY] * self._num_chans
        scale = [CHARTSCALE] * self._num_chans

        # get X axis (time) range for all channels and draw labels
        mintim, maxtim = self._mintim, self._maxtim
        mintim = maxtim - float(self._timrange.get())
        bounds = AreaXY(mintim, CHARTMINY, maxtim, CHARTMAXY)
        self._can_chartview.delete(AXISLBLTAG)
        self._draw_xaxis_labels(w, h, bounds, mintim, maxtim)
        self._spn_timrange.configure(fg=LBLCOL, readonlybackground=BGCOL)

        # plot each channel's data points
        for chn in range(self._num_chans):

            chncol = self._plotcols[chn]
            minval[chn] = float(self._data_miny[chn].get())
            maxval[chn] = float(self._data_maxy[chn].get())
            scale[chn] = float(self._data_scale[chn].get())
            bounds = AreaXY(mintim, minval[chn], maxtim, maxval[chn])

            # draw Y axis (data value) labels for this channel
            self._draw_yaxis_labels(w, h, bounds, minval[chn], maxval[chn], chn)

            # plot each data point in channel
            inr = False
            for tim, channels in data.items():
                try:
                    val = channels[chn]
                except KeyError:
                    val = None

                if val is None:
                    continue

                if scale[chn] != 1:
                    val /= scale[chn]  # scale data

                # convert datapoint to canvas x,y coordinates
                x, y = data2xy(
                    w - self._xoff * 2,
                    h - self._yoff,
                    bounds,
                    tim,
                    val,
                    self._xoff,
                )
                if x <= self._xoff:
                    inr = False
                # plot line
                if inr:
                    x2, y2 = x, y
                    self._can_chartview.create_line(
                        x1,
                        y1,
                        x2,
                        y2,
                        fill=chncol,
                        width=PLOTWID,
                        tags=f"plot_{chn:1d}",
                    )
                    x1, y1 = x2, y2
                else:
                    x1, y1 = max(x, self._xoff), y
                inr = True

    def _draw_xaxis_labels(
        self, w: int, h: int, bounds: AreaXY, mintim: float, maxtim: float
    ):
        """
        Draw X axis (time) labels.

        :param int w: canvas width
        :param int h: canvas height
        :param AreaXY bounds: data bounds
        :param float mintim: minimum time
        :param float maxtim: maximum time
        """
        # pylint: disable=too-many-arguments, too-many-positional-arguments

        for g in XLBLSTEPS:
            xval = mintim + (maxtim - mintim) * g
            x, _ = data2xy(w - self._xoff * 2, h - self._yoff, bounds, xval, 0)
            if g == 0:
                anc = NW
            elif g == 1:
                anc = NE
            else:
                anc = N
            self._can_chartview.create_text(
                x + self._xoff,
                h - self._yoff,
                text=time2str(xval),
                anchor=anc,
                fill=AXISCOL,
                font=self._font,
                tags=AXISLBLTAG,
            )

    def _draw_yaxis_labels(
        self,
        w: int,
        h: int,
        bounds: AreaXY,
        minval: float,
        maxval: float,
        chn: int,
    ):
        """
        Draw Y axis (data value) labels for this channel.

        :param int w: canvas width
        :param int h: canvas height
        :param AreaXY bounds: data bounds
        :param float minval: minimum val for chn
        :param float maxval: maximum val for chn
        :param int chn: channel
        """
        # pylint: disable=too-many-arguments, too-many-positional-arguments

        col = self._plotcols[chn]
        yo = 2  # avoid edges
        # y axis labels alternate left and right
        if chn % 2:  # odd channels
            x = w - yo - self._fonth * ((chn - 1) / 2)
        else:  # even channels
            x = yo + self._fonth * (chn / 2)

        for g in YLBLSTEPS:
            yval = minval + (maxval - minval) * g
            _, y = data2xy(
                w - self._xoff * MAXCHANS / 2, h - self._yoff, bounds, 0, yval
            )
            if g == 0:
                anc = SW if chn % 2 else NW
            elif g == 1:
                y += yo * 2  # avoid edges
                anc = SE if chn % 2 else NE
            else:
                anc = S if chn % 2 else N
            self._can_chartview.create_text(
                x,
                y,
                text=yval,
                fill=col,
                font=self._font,
                angle=90,
                anchor=anc,
                tags=AXISLBLTAG,
            )

    def _disp_alert(self, msg):
        """
        Display alert message on canvas.

        :param str msg: error message
        """

        w, h = self.width, self.height

        self._can_chartview.delete(ALL)
        self._can_chartview.create_text(
            w / 2,
            h / 2,
            text=msg,
            fill="orange",
            font=self._font,
            anchor=S,
        )

    def _on_clipboard(self, event):  # pylint: disable=unused-argument
        """
        Copy chart data to clipboard in CSV format.

        :param event event: double click event
        """

        csv = (
            f"PyGPSClient Chart Data,{time2str(time(),'%Y-%m-%d-%H:%M:%S')},"
            f"Channels,{self._num_chans}\n"
        )
        hdr = True
        for tim, data in self._chart_data.items():
            if hdr:
                csv += "Timestamp"
                for chn in range(self._num_chans):
                    csv += f",{self._ent_id[chn].get()}.{self._ent_name[chn].get()}"
                csv += "\n"
                hdr = False
            csv += f"{time2str(tim,'%Y-%m-%d-%H:%M:%S')}"
            for chn in range(self._num_chans):
                try:
                    csv += f",{data[chn]}"
                except KeyError:
                    csv += ","
            csv += "\n"
        self.__master.clipboard_clear()
        self.__master.clipboard_append(csv)
        self.__master.update()

    def _on_resize(self, event):  # pylint: disable=unused-argument
        """
        Resize frame.

        :param event event: resize event
        """

        self.width, self.height = self.get_size()
        self._font, self._fonth = scale_font(self.width, 6, 35, 16)

    def get_size(self):
        """
        Get current canvas size.

        :return: window size (width, height)
        :rtype: tuple
        """

        self.update_idletasks()  # Make sure we know about any resizing
        return self._can_chartview.winfo_width(), self._can_chartview.winfo_height()



"""
globals.py

PyGPSClient Globals

Collection of global constants

Created on 14 Sep 2020

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

# pylint: disable=line-too-long

from collections import namedtuple
from datetime import datetime
from os import path
from pathlib import Path
from tkinter import Canvas

from pyubx2 import GET, POLL, SET, SETPOLL

Point = namedtuple("Point", ["lat", "lon"])
# Area convention is minlat, minlon, maxlat, maxlon
Area = namedtuple("Area", ["lat1", "lon1", "lat2", "lon2"])
TrackPoint = namedtuple("TrackPoint", ["lat", "lon", "tim", "ele", "spd"])
PointXY = namedtuple("PointXY", ["x", "y"])
AreaXY = namedtuple("AreaXY", ["x1", "y1", "x2", "y2"])


def create_circle(self, x, y, r, **kwargs):
    """
    Helper method to simplify drawing circles on canvas

    :param int x: x coordinate
    :param int y: y coordinate
    :param int r: radius
    """

    return self.create_oval(x - r, y - r, x + r, y + r, **kwargs)


Canvas.create_circle = create_circle

HOME = Path.home()
APPNAME = __name__.split(".", 1)[0]  # i.e. "pygpsclient"
ASCII = "ascii"
AXISCOL = "white"
BGCOL = "gray24"  # default widget background color
BPSRATES = (
    9600,
    19200,
    38400,
    57600,
    115200,
    230400,
    460800,
    921600,
    1000000,
    4800,
)
BSR = "backslashreplace"
CFG = "cfg"
CLASS = "cls"
COLORTAGS = "colortags"
CONFIGFILE = path.join(HOME, f"{APPNAME}.json")
CONNECTED = 1
CONNECTED_FILE = 4
CONNECTED_NTRIP = 8
CONNECTED_SIMULATOR = 32
CONNECTED_SOCKET = 2
CONNECTED_SPARTNIP = 16
CONNECTED_SPARTNLB = CONNECTED
CONNMODES = {
    0: "not connected",
    1: "serial",
    2: "file",
    3: "no serial devices",
    4: "socket",
}
CRLF = b"\x0d\x0a"
DDD = "DD.D"
DEFAULT_BUFSIZE = 4096
DEFAULT_PASSWORD = "password"  # nosec
DEFAULT_PORT = 50010
DEFAULT_TLS_PORTS = (443, 2102)
DEFAULT_REGION = "eu"
DEFAULT_SERVER = "localhost"
DEFAULT_USER = "anon"
DIRNAME = path.dirname(__file__)
DISCONNECTED = 0
DMM = "DM.M"
DMS = "D.M.S"
ECEF = "ECEF"
ENABLE_CFG_LEGACY = True  # enable CFG=* Other Configuration command panel
ERRCOL = "salmon"  # default invalid data entry field background color
ERROR = "ERR!"
FGCOL = "white"  # default widget foreground color
FILEREAD_INTERVAL = 0.02  # delay between successive datalog file reads (seconds)
FONT_FIXED = "TkFixedFont"
FONT_MENU = "TkMenuFont"
FONT_TEXT = "TkTextFont"
FORMAT_BINARY = "Binary"
FORMAT_BOTH = "Parsed + Hex Tabular"
FORMAT_HEXSTR = "Hex String"
FORMAT_HEXTAB = "Hex Tabular"
FORMAT_PARSED = "Parsed"
FORMATS = (
    FORMAT_PARSED,
    FORMAT_BINARY,
    FORMAT_HEXSTR,
    FORMAT_HEXTAB,
    FORMAT_BOTH,
)
FRAME = "frm"
GGA_INTERVALS = ("None", "2", "5", "10", "60", "120")
GITHUB_URL = "https://github.com/semuconsulting/PyGPSClient"
GLONASS_NMEA = True  # use GLONASS NMEA SVID (65-96) rather than slot (1-24)
GNSS = "GNSS"
GNSS_EOF_EVENT = "<<gnss_eof>>"
GNSS_ERR_EVENT = "<<gnss_error>>"
GNSS_EVENT = "<<gnss_read>>"
GNSS_TIMEOUT_EVENT = "<<gnss_timeout>>"
GNSS_LIST = {
    0: ("GPS", "cornflowerblue"),
    1: ("SBA", "orange"),
    2: ("GAL", "green4"),
    3: ("BEI", "mediumpurple2"),
    4: ("IME", "violet"),
    5: ("QZS", "yellow"),
    6: ("GLO", "indianred"),
    7: ("NAV", "grey60"),
}
GPSEPOCH0 = datetime(1980, 1, 6)  # for Wno and Tow calculations
GPX_NS = (
    'xmlns="http://www.topografix.com/GPX/1/1" '
    'creator="PyGPSClient" version="1.1" '
    'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
    'xsi:schemaLocation="http://www.topografix.com/GPX/1/1 '
    'http://www.topografix.com/GPX/1/1/gpx.xsd"'
)
GPX_TRACK_INTERVAL = 1  # minimum GPS track update interval (seconds)
GRIDCOL = "grey40"
GUI_UPDATE_INTERVAL = 0.5  # GUI widget update interval (seconds)
ICON_APP128 = path.join(DIRNAME, "resources/app-128.png")
ICON_BLANK = path.join(DIRNAME, "resources/blank-1-24.png")
ICON_CONFIRMED = path.join(DIRNAME, "resources/iconmonstr-check-mark-8-24.png")
ICON_CONN = path.join(DIRNAME, "resources/iconmonstr-media-control-48-24.png")
ICON_CONTRACT = path.join(DIRNAME, "resources/iconmonstr-triangle-1-16.png")
ICON_DELETE = path.join(DIRNAME, "resources/iconmonstr-trash-can-filled-24.png")
ICON_DISCONN = path.join(DIRNAME, "resources/iconmonstr-media-control-50-24.png")
ICON_END = path.join(DIRNAME, "resources/marker_end.png")
ICON_EXIT = path.join(DIRNAME, "resources/iconmonstr-door-6-24.png")
ICON_EXPAND = path.join(DIRNAME, "resources/iconmonstr-arrow-80-16.png")
ICON_GITHUB = path.join(DIRNAME, "resources/github-256.png")
ICON_LOAD = path.join(DIRNAME, "resources/iconmonstr-folder-18-24.png")
ICON_LOGREAD = path.join(DIRNAME, "resources/binary-1-24.png")
ICON_NMEACONFIG = path.join(DIRNAME, "resources/iconmonstr-gear-2-24-brown.png")
ICON_NOCLIENT = path.join(DIRNAME, "resources/iconmonstr-noclient-10-24.png")
ICON_NOTRANSMIT = path.join(DIRNAME, "resources/iconmonstr-notransmit-10-24.png")
ICON_NTRIPCONFIG = path.join(DIRNAME, "resources/iconmonstr-antenna-4-24.png")
ICON_PENDING = path.join(DIRNAME, "resources/iconmonstr-time-6-24.png")
ICON_PLAY = path.join(DIRNAME, "resources/iconmonstr-media-control-48-24.png")
ICON_POS = path.join(DIRNAME, "resources/iconmonstr-plus-lined-24.png")
ICON_RECORD = path.join(DIRNAME, "resources/iconmonstr-record-24.png")
ICON_REDRAW = path.join(DIRNAME, "resources/iconmonstr-refresh-lined-24.png")
ICON_REFRESH = path.join(DIRNAME, "resources/iconmonstr-refresh-6-16.png")
ICON_SAVE = path.join(DIRNAME, "resources/iconmonstr-save-14-24.png")
ICON_SEND = path.join(DIRNAME, "resources/iconmonstr-arrow-12-24.png")
ICON_SERIAL = path.join(DIRNAME, "resources/usbport-1-24.png")
ICON_SOCKET = path.join(DIRNAME, "resources/ethernet-1-24.png")
ICON_SPARTNCONFIG = path.join(DIRNAME, "resources/iconmonstr-antenna-3-24.png")
ICON_START = path.join(DIRNAME, "resources/marker_start.png")
ICON_STOP = path.join(DIRNAME, "resources/iconmonstr-stop-1-24.png")
ICON_TRANSMIT = path.join(DIRNAME, "resources/iconmonstr-transmit-10-24.png")
ICON_UBXCONFIG = path.join(DIRNAME, "resources/iconmonstr-gear-2-24.png")
ICON_UNDO = path.join(DIRNAME, "resources/iconmonstr-undo-24.png")
ICON_UNKNOWN = path.join(DIRNAME, "resources/clear-1-24.png")
ICON_WARNING = path.join(DIRNAME, "resources/iconmonstr-warning-1-24.png")
IMG_WORLD = path.join(DIRNAME, "resources/world.png")
ICON_SPONSOR = path.join(DIRNAME, "resources/bmc-full-logo-no-background.png")
IMG_WORLD_BOUNDS = Area(-90, -180, 90, 180)
INFOCOL = "steelblue2"

KNOWNGPS = (
    "CP210",
    "FT230",
    "FT232",
    "garmin",
    "gnss",
    "gps",
    "magellan",
    "navstar",
    "septentrio",
    "sirf",
    "trimble",
    "u-blox",
    "ublox",
    "USB Serial",
    "USB to UART",
    "USB UART",
    "USB_to_UART",
    "USER-DEFINED",
)
LBAND = "LBAND"
LICENSE_URL = "https://github.com/semuconsulting/PyGPSClient/blob/master/LICENSE"
M2FT = 3.28084
MAPAPI_URL = "https://developer.mapquest.com/user/login/sign-up"
MQTTIPMODE = 0
MQTTLBANDMODE = 1
MINHEIGHT = 600
MINWIDTH = 800
CUSTOM = "custom"
IMPORT = "import"
WORLD = "world"
LG290P = "Quectel LG290P"
MAX_SNR = 60  # upper limit of graphview snr axis
MAXLOGLINES = 10000  # maximum number of 'lines' per datalog file
MIN_GUI_UPDATE_INTERVAL = 0.1  # minimum GUI widget update interval (seconds)
MOSAIC_X5 = "Septentrio Mosaic X5"
MQAPIKEY = "mqapikey"
MQTT_PROTOCOL = 16
MSGMODES = {
    "GET": GET,
    "SET": SET,
    "POLL": POLL,
    "SETPOLL": SETPOLL,
}
NOPORTS = 3
NTRIP = "NTRIP"
NTRIP_EVENT = "<<ntrip_read>>"
OKCOL = "green"
PASSTHRU = "Passthrough"
PNTCOL = "orange"
PORTIDS = ("0 I2C", "1 UART1", "2 UART2", "3 USB", "4 SPI")
PUBLICIP_URL = "https://ipinfo.io/json"
PYPI_URL = "https://pypi.org/pypi/PyGPSClient"
QUITONERRORDEFAULT = 1
RCVR_CONNECTION = "USB,UART1"  # default GNSS receiver connection port(s)
ROMVER_NEW = "23.01"  # min device ROM version using configuration database
READONLY = "readonly"
RESIZE = "resize"
ROUTE = "route"
RXMMSG = "RXM-SPARTN-KEY"
SAT_EXPIRY = 10  # how long passed satellites are kept in the sky and graph view
SBF_PROTOCOL = 64
SCREENSCALE = 0.8  # screen resolution scaling factor
SOCK_NTRIP = "NTRIP CASTER"
SOCK_SERVER = "SOCKET SERVER"
SOCKCLIENT_HOST = "localhost"
SOCKCLIENT_PORT = 50010
SOCKMODES = (SOCK_SERVER, SOCK_NTRIP)
SOCKSERVER_HOST = "0.0.0.0"  # i.e. bind to all host IP addresses
SOCKSERVER_MAX_CLIENTS = 5
SOCKSERVER_NTRIP_PORT = 2101
SOCKSERVER_PORT = 50012
SPARTN_EOF_EVENT = "<<spartn_eof>>"
SPARTN_ERR_EVENT = "<<spartn_error>>"
SPARTN_EVENT = "<<spartn_read>>"
SPARTN_KEYLEN = 16
SPARTN_DEFAULT_KEY = "abcd1234abcd1234abcd1234abcd1234"
SPARTN_BASEDATE_CURRENT = -1
SPARTN_BASEDATE_DATASTREAM = 0
SPARTN_OUTPORT = 8883
SPARTN_PPREGIONS = ("eu", "us", "jp", "kr", "au")
SPARTN_PPSERVER_URL = "pp.services.u-blox.com"
SPARTN_PROTOCOL = 8
SPARTN_SOURCE_IP = 0
SPARTN_SOURCE_LB = 1
SPONSOR_URL = "https://buymeacoffee.com/semuconsulting"
SQRT2 = 0.7071067811865476  # square root of 2
THD = "thd"
TIME0 = datetime(1970, 1, 1)  # basedate for time()
TIMEOUTS = (
    "0.1",
    "0.2",
    "1",
    "2",
    "5",
    "10",
    "20",
    "None",
    "0",
)
TOPIC_IP = "/pp/ip/{}"
TOPIC_MGA = "/pp/ubx/mga"
TOPIC_RXM = "/pp/ubx/0236/ip"
TRACK = "track"
TTY_EVENT = "<<tty_read>>"
TTY_PROTOCOL = 32
TTYOK = ("OK", "$R:")
TTYERR = ("ERROR", "$R?")
TTYMARKER = "TTY<<"
UBXPRESETS = "ubxpresets"
UBXSIMULATOR = "ubxsimulator"
UTF8 = "utf-8"
WAYPOINT = "waypoint"
WIDGETU1 = (200, 200)  # small widget size
WIDGETU2 = (300, 200)  # medium widget size
WIDGETU3 = (800, 200)  # Console size
WIDGETU4 = (500, 500)  # GPX Track viewer size
WIDGETU6 = (400, 200)  # Chart size
XML_HDR = '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
ZED_F9 = "u-blox ZED-F9/X20"

# Conversion factors
UI = "Imperial mph"
UIK = "Imperial knots"
UMK = "Metric kmph"
UMM = "Metric m/s"
KM2M = 1000
KM2MIL = 0.621371
KM2NMIL = 0.5399568
KPH2KNT = 0.5399568
KPH2MPH = 0.6213712
KPH2MPS = 0.2777778
M2MIL = 0.0006213712
M2NMIL = 0.0005399568
M2KM = 0.001
MPS2MPH = 2.236936
MPS2KPH = 3.6
MPS2KNT = 1.943844

# UBX & NMEA config widget signifiers - used
# to identify which widget should receive the
# data from a given POLL or ACK message:
UBX_MONVER = 0
UBX_MONHW = 1
UBX_CFGPRT = 2
UBX_CFGMSG = 3
UBX_CFGVAL = 4
UBX_PRESET = 5
UBX_CFGRATE = 6
UBX_CFGOTHER = 7
SPARTN_GNSS = 8
SPARTN_LBAND = 9
SPARTN_MQTT = 10
SPECTRUMVIEW = 11
SYSMONVIEW = 12
ROVERVIEW = 13
UBX_MONRF = 14
NMEA_MONHW = 15
NMEA_PRESET = 16
NMEA_CFGOTHER = 17
SERVERCONFIG = 18
SBF_MONHW = 19

# map of fix values to descriptions
# the keys in this map are a concatenation of NMEA/UBX
# message identifier and attribute value e.g.:
# GGA1: GGA + quality = 1
# NAV-STATUS3: NAV-STATUS + gpsFix = 3
# (valid for NMEA >=4)
FIXLOOKUP = {
    "GGA1": "3D",  # quality
    "GGA2": "3D",
    "GGA4": "RTK FIXED",
    "GGA5": "RTK FLOAT",
    "GGA6": "DR",
    "GLLA": "3D",  # posMode
    "GLLD": "RTK",
    "GLLE": "DR",
    "GNSA": "3D",  # posMode
    "GNSD": "3D",
    "GNSF": "RTK FLOAT",
    "GNSR": "RTK FIXED",
    "GNSE": "DR",
    "RMCA": "3D",  # posMode
    "RMCD": "3D",
    "RMCF": "RTK FLOAT",
    "RMCR": "RTK FIXED",
    "RMCE": "DR",
    "VTGA": "3D",  # posMode
    "VTGD": "RTK",
    "VTGE": "DR",
    "NAV-PVT1": "DR",  # fixType
    "NAV-PVT2": "2D",
    "NAV-PVT3": "3D",
    "NAV-PVT4": "GNSS+DR",
    "NAV-PVT5": "TIME ONLY",
    "NAV-PVT6": "RTK FLOAT",  # carrSoln
    "NAV-PVT7": "RTK FIXED",
    "NAV-STATUS1": "DR",  # gpsFix
    "NAV-STATUS2": "3D",
    "NAV-STATUS3": "3D",
    "NAV-STATUS4": "GNSS+DR",
    "NAV-STATUS5": "TIME ONLY",
    "NAV-STATUS6": "RTK FLOAT",  # carrSoln
    "NAV-STATUS7": "RTK FIXED",
    "NAV-SOL1": "DR",  # gpsFix
    "NAV-SOL2": "3D",
    "NAV-SOL3": "3D",
    "NAV-SOL4": "GNSS+DR",
    "NAV-SOL5": "TIME ONLY",
    "HNR-PVT1": "DR",  # gpsFix
    "HNR-PVT2": "2D",
    "HNR-PVT3": "3D",
    "HNR-PVT4": "GNSS+DR",
    "HNR-PVT5": "TIME ONLY",
    "HNR-PVT6": "RTK",
    "NAV2-PVT1": "DR",  # fixType
    "NAV2-PVT2": "2D",
    "NAV2-PVT3": "3D",
    "NAV2-PVT4": "GNSS+DR",
    "NAV2-PVT5": "TIME ONLY",
    "NAV2-PVT6": "RTK FLOAT",  # carrSoln
    "NAV2-PVT7": "RTK FIXED",
    "NAV2-STATUS1": "DR",  # gpsFix
    "NAV2-STATUS2": "3D",
    "NAV2-STATUS3": "3D",
    "NAV2-STATUS4": "GNSS+DR",
    "NAV2-STATUS5": "TIME ONLY",
    "NAV2-STATUS6": "RTK FLOAT",  # carrSoln
    "NAV2-STATUS7": "RTK FIXED",
    "PVTGeodetic0": "NO FIX",
    "PVTGeodetic1": "3D",
    "PVTGeodetic2": "RTK",
    "PVTGeodetic3": "3D",
    "PVTGeodetic4": "RTK FIXED",
    "PVTGeodetic5": "RTK FLOAT",
    "PVTGeodetic6": "SBAS",
    "PVTGeodetic7": "RTK FIXED",
    "PVTGeodetic8": "RTK FLOAT",
    "PVTGeodetic10": "PPP",
}



"""
ubx_port_frame.py

UBX Configuration frame for CFG-PRT commands

Created on 22 Dec 2020

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

from tkinter import Button, Checkbutton, E, Frame, IntVar, Label, Spinbox, StringVar, W

from PIL import Image, ImageTk
from pyubx2 import POLL, SET, UBXMessage

from pygpsclient.globals import (
    BPSRATES,
    CONNECTED,
    ERRCOL,
    ICON_CONFIRMED,
    ICON_PENDING,
    ICON_SEND,
    ICON_WARNING,
    OKCOL,
    PORTIDS,
    READONLY,
    UBX_CFGPRT,
)
from pygpsclient.strings import LBLCFGPRT


class UBX_PORT_Frame(Frame):
    """
    UBX Port and Protocol configuration command panel.
    """

    def __init__(self, app, container, *args, **kwargs):
        """
        Constructor.

        :param Frame app: reference to main tkinter application
        :param Frame container: reference to container frame (config-dialog)
        :param args: optional args to pass to Frame parent class
        :param kwargs: optional kwargs to pass to Frame parent class
        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.appmaster  # Reference to root class (Tk)
        self.__container = container

        super().__init__(container.container, *args, **kwargs)

        self._img_send = ImageTk.PhotoImage(Image.open(ICON_SEND))
        self._img_pending = ImageTk.PhotoImage(Image.open(ICON_PENDING))
        self._img_confirmed = ImageTk.PhotoImage(Image.open(ICON_CONFIRMED))
        self._img_warn = ImageTk.PhotoImage(Image.open(ICON_WARNING))
        self._bpsrate = IntVar()
        self._portid = StringVar()
        self._inprot = (1, 1, 0, 1)
        self._outprot = (1, 1, 0)
        self._inprot_nmea = IntVar()
        self._inprot_ubx = IntVar()
        self._inprot_rtcm2 = IntVar()
        self._inprot_rtcm3 = IntVar()
        self._outprot_nmea = IntVar()
        self._outprot_ubx = IntVar()
        self._outprot_rtcm3 = IntVar()

        self._body()
        self._do_layout()
        self._attach_events()

    def _body(self):
        """
        Set up frame and widgets.
        """

        self._lbl_cfg_port = Label(self, text=LBLCFGPRT, anchor=W)
        self._lbl_ubx_portid = Label(self, text="Port ID")
        self._spn_ubx_portid = Spinbox(
            self,
            values=PORTIDS,
            width=8,
            state=READONLY,
            wrap=True,
            textvariable=self._portid,
            command=lambda: self._on_select_portid(),  # pylint: disable=unnecessary-lambda
        )
        self._lbl_ubx_bpsrate = Label(self, text="Rate bps")
        self._spn_ubx_bpsrate = Spinbox(
            self,
            values=(BPSRATES),
            width=6,
            state=READONLY,
            wrap=True,
            textvariable=self._bpsrate,
        )
        self._lbl_inprot = Label(self, text="Input")
        self._chk_inprot_nmea = Checkbutton(
            self, text="NMEA", variable=self._inprot_nmea
        )
        self._chk_inprot_ubx = Checkbutton(self, text="UBX", variable=self._inprot_ubx)
        self._chk_inprot_rtcm2 = Checkbutton(
            self, text="RTCM2", variable=self._inprot_rtcm2
        )
        self._chk_inprot_rtcm3 = Checkbutton(
            self, text="RTCM3", variable=self._inprot_rtcm3
        )
        self._lbl_outprot = Label(self, text="Output")
        self._chk_outprot_nmea = Checkbutton(
            self, text="NMEA", variable=self._outprot_nmea
        )
        self._chk_outprot_ubx = Checkbutton(
            self, text="UBX", variable=self._outprot_ubx
        )
        self._chk_outprot_rtcm3 = Checkbutton(
            self, text="RTCM3", variable=self._outprot_rtcm3
        )
        self._lbl_send_command = Label(self, image=self._img_pending)
        self._btn_send_command = Button(
            self,
            image=self._img_send,
            width=50,
            command=self._on_send_port,
            font=self.__app.font_md,
        )

    def _do_layout(self):
        """
        Layout widgets.
        """

        self._lbl_cfg_port.grid(column=0, row=0, columnspan=6, padx=3, sticky=(W, E))
        self._lbl_ubx_portid.grid(
            column=0, row=1, columnspan=1, rowspan=2, padx=3, sticky=W
        )
        self._spn_ubx_portid.grid(column=1, row=1, columnspan=1, rowspan=2, sticky=W)
        self._lbl_ubx_bpsrate.grid(
            column=2, row=1, columnspan=1, rowspan=2, padx=3, sticky=W
        )
        self._spn_ubx_bpsrate.grid(column=3, row=1, columnspan=2, rowspan=2, sticky=W)
        self._lbl_inprot.grid(column=0, row=3, padx=3, sticky=W)
        self._chk_inprot_nmea.grid(column=1, row=3, sticky=W)
        self._chk_inprot_ubx.grid(column=2, row=3, sticky=W)
        self._chk_inprot_rtcm2.grid(column=3, row=3, sticky=W)
        self._chk_inprot_rtcm3.grid(column=4, row=3, sticky=W)
        self._lbl_outprot.grid(column=0, row=4, padx=3, sticky=W)
        self._chk_outprot_nmea.grid(column=1, row=4, sticky=W)
        self._chk_outprot_ubx.grid(column=2, row=4, sticky=W)
        self._chk_outprot_rtcm3.grid(column=3, row=4, sticky=W)
        self._btn_send_command.grid(
            column=4, row=1, rowspan=2, ipadx=3, ipady=3, sticky=E
        )
        self._lbl_send_command.grid(
            column=5, row=1, rowspan=2, ipadx=3, ipady=3, sticky=E
        )

        (cols, rows) = self.grid_size()
        for i in range(cols):
            self.grid_columnconfigure(i, weight=1)
        for i in range(rows):
            self.grid_rowconfigure(i, weight=1)
        self.option_add("*Font", self.__app.font_sm)

    def _attach_events(self):
        """
        Bind events to widget.
        """

        # click mouse button to refresh information
        self.bind("<Button>", self._do_poll_prt)

    def reset(self):
        """
        Reset panel to initial settings
        """

        if self.__app.conn_status == CONNECTED:
            self._do_poll_prt()

    def update_status(self, msg: UBXMessage):
        """
        Update pending confirmation status.

        :param UBXMessage msg: UBX config message
        """

        if msg.identity == "CFG-PRT":
            self._bpsrate.set(msg.baudRate)
            self._inprot_ubx.set(msg.inUBX)
            self._inprot_nmea.set(msg.inNMEA)
            self._inprot_rtcm2.set(msg.inRTCM)
            self._inprot_rtcm3.set(msg.inRTCM3)
            self._outprot_ubx.set(msg.outUBX)
            self._outprot_nmea.set(msg.outNMEA)
            self._outprot_rtcm3.set(msg.outRTCM3)
            self._lbl_send_command.config(image=self._img_confirmed)
            self.__container.set_status("CFG-PRT GET message received", OKCOL)

        elif msg.identity == "ACK-NAK":
            self.__container.set_status("CFG-PRT POLL message rejected", ERRCOL)
            self._lbl_send_command.config(image=self._img_warn)

    def _on_select_portid(self):
        """
        Handle portid selection.
        """

        self._do_poll_prt()

    def _on_send_port(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Handle Send port config button press.
        """

        portID = int(self._portid.get()[0:1])
        baudRate = int(self._bpsrate.get())
        inUBX = self._inprot_ubx.get()
        inNMEA = self._inprot_nmea.get()
        inRTCM = self._inprot_rtcm2.get()
        inRTCM3 = self._inprot_rtcm3.get()
        outUBX = self._outprot_ubx.get()
        outNMEA = self._outprot_nmea.get()
        outRTCM3 = self._outprot_rtcm3.get()
        msg = UBXMessage(
            "CFG",
            "CFG-PRT",
            SET,
            portID=portID,
            charLen=3,  # 8 data bits
            parity=4,  # none
            nStopBits=0,  # 1 stop bit
            baudRate=baudRate,
            inUBX=inUBX,
            inNMEA=inNMEA,
            inRTCM=inRTCM,
            inRTCM3=inRTCM3,
            outUBX=outUBX,
            outNMEA=outNMEA,
            outRTCM3=outRTCM3,
        )
        self.__container.send_command(msg)
        self._lbl_send_command.config(image=self._img_pending)
        self.__container.set_status(
            "CFG-PRT SET message sent",
        )
        for msgid in ("ACK-NAK", "ACK-NAK"):
            self.__container.set_pending(msgid, UBX_CFGPRT)

        self._do_poll_prt()

    def _do_poll_prt(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Poll PRT message configuration.
        """

        portID = int(self._portid.get()[0:1])
        msg = UBXMessage("CFG", "CFG-PRT", POLL, portID=portID)
        self.__container.send_command(msg)
        self._lbl_send_command.config(image=self._img_pending)
        self.__container.set_status(
            "CFG-PRT POLL message sent",
        )
        for msgid in ("CFG-PRT", "ACK-NAK"):
            self.__container.set_pending(msgid, UBX_CFGPRT)



"""
settings_frame.py

Settings frame class for PyGPSClient application.

- Holds all the latest settings in self.config
- Sets initial (saved) configuration of the following frames:
- frm_settings (SettingsFrame class) for general application settings.
- frm_serial (SerialConfigFrame class) for serial port settings.
- frm_socketclient (SocketConfigFrame class) for socket client settings.
- frm_socketserver (ServerConfigFrame class) for socket server settings.

Created on 12 Sep 2020

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

# pylint: disable=unnecessary-lambda

from platform import system
from tkinter import (
    ALL,
    BOTH,
    BOTTOM,
    DISABLED,
    HORIZONTAL,
    LEFT,
    NORMAL,
    NW,
    RIGHT,
    VERTICAL,
    Button,
    Canvas,
    Checkbutton,
    E,
    Frame,
    IntVar,
    Label,
    Scrollbar,
    Spinbox,
    StringVar,
    TclError,
    W,
    X,
    Y,
    ttk,
)

from PIL import Image, ImageTk
from pyubx2 import NMEA_PROTOCOL, RTCM3_PROTOCOL, UBX_PROTOCOL

from pygpsclient.globals import (
    BPSRATES,
    CONNECTED,
    CONNECTED_FILE,
    CONNECTED_SOCKET,
    DDD,
    DISCONNECTED,
    DMM,
    DMS,
    ECEF,
    ERRCOL,
    FORMATS,
    GNSS,
    GNSS_EOF_EVENT,
    GNSS_ERR_EVENT,
    GNSS_EVENT,
    GNSS_TIMEOUT_EVENT,
    HOME,
    ICON_CONN,
    ICON_DISCONN,
    ICON_EXIT,
    ICON_LOGREAD,
    ICON_NMEACONFIG,
    ICON_NTRIPCONFIG,
    ICON_SERIAL,
    ICON_SOCKET,
    ICON_SPARTNCONFIG,
    ICON_UBXCONFIG,
    KNOWNGPS,
    MSGMODES,
    NOPORTS,
    OKCOL,
    READONLY,
    SBF_PROTOCOL,
    SPARTN_PROTOCOL,
    TIMEOUTS,
    TTY_PROTOCOL,
    UI,
    UIK,
    UMK,
    UMM,
)
from pygpsclient.helpers import fontheight, fontwidth
from pygpsclient.serialconfig_frame import SerialConfigFrame
from pygpsclient.serverconfig_frame import ServerConfigFrame
from pygpsclient.socketconfig_frame import SocketConfigFrame
from pygpsclient.sqllite_handler import SQLOK
from pygpsclient.strings import (
    DLGTNMEA,
    DLGTNTRIP,
    DLGTSPARTN,
    DLGTUBX,
    LBLDATABASERECORD,
    LBLDATADISP,
    LBLDATALOG,
    LBLDEGFORMAT,
    LBLNMEACONFIG,
    LBLNTRIPCONFIG,
    LBLPROTDISP,
    LBLSHOWUNUSED,
    LBLSPARTNCONFIG,
    LBLTRACKRECORD,
    LBLUBXCONFIG,
)

MAXLINES = ("200", "500", "1000", "2000", "100")
# initial dimensions adjusted for different widget
# rendering on different platforms
if system() == "Linux":  # Wayland
    MINHEIGHT = 28
    MINWIDTH = 28
elif system() == "Darwin":  # MacOS

    MINHEIGHT = 38
    MINWIDTH = 30
else:  # Windows and others
    MINHEIGHT = 35
    MINWIDTH = 26


class SettingsFrame(Frame):
    """
    Settings frame class.
    """

    def __init__(self, app, *args, **kwargs):
        """
        Constructor.

        :param Frame app: reference to main tkinter application
        :param args: optional args to pass to Frame parent class
        :param kwargs: optional kwargs to pass to Frame parent class
        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.appmaster  # Reference to root class (Tk)

        Frame.__init__(self, self.__master, *args, **kwargs)

        self.infilepath = None
        self.logpath = HOME
        self.trackpath = HOME
        self.databasepath = HOME
        self._prot_nmea = IntVar()
        self._prot_ubx = IntVar()
        self._prot_sbf = IntVar()
        self._prot_rtcm3 = IntVar()
        self._prot_spartn = IntVar()
        self._prot_tty = IntVar()
        self._autoscroll = IntVar()
        self._maxlines = IntVar()
        self._units = StringVar()
        self._degrees_format = StringVar()
        self._console_format = StringVar()
        self._datalog = IntVar()
        self._logformat = StringVar()
        self._record_track = IntVar()
        self._record_database = IntVar()
        self._show_unusedsat = IntVar()
        self.show_legend = IntVar()
        self._colortag = IntVar()
        self.defaultports = self.__app.configuration.get("defaultport_s")
        self._validsettings = True
        self._img_conn = ImageTk.PhotoImage(Image.open(ICON_CONN))
        self._img_serial = ImageTk.PhotoImage(Image.open(ICON_SERIAL))
        self._img_socket = ImageTk.PhotoImage(Image.open(ICON_SOCKET))
        self._img_disconn = ImageTk.PhotoImage(Image.open(ICON_DISCONN))
        self._img_exit = ImageTk.PhotoImage(Image.open(ICON_EXIT))
        self._img_ubxconfig = ImageTk.PhotoImage(Image.open(ICON_UBXCONFIG))
        self._img_nmeaconfig = ImageTk.PhotoImage(Image.open(ICON_NMEACONFIG))
        self._img_ntripconfig = ImageTk.PhotoImage(Image.open(ICON_NTRIPCONFIG))
        self._img_spartnconfig = ImageTk.PhotoImage(Image.open(ICON_SPARTNCONFIG))
        self._img_dataread = ImageTk.PhotoImage(Image.open(ICON_LOGREAD))

        self._container()  # create scrollable container
        self._body()
        self._do_layout()
        self.reset()

    def _container(self):
        """
        Create scrollable container frame.

        NB: any expandable sub-frames must implement an on_resize()
        function which invokes the on_expand() method here.
        """

        dimw = fontwidth(self.__app.font_md) * MINWIDTH
        dimh = fontheight(self.__app.font_md) * MINHEIGHT
        self._frm_main = Frame(self)
        self._frm_main.pack(fill=BOTH, expand=1)
        self_frm_scrollx = Frame(self._frm_main)
        self_frm_scrollx.pack(fill=X, side=BOTTOM)
        self._can_container = Canvas(self._frm_main, height=dimh, width=dimw)
        self._frm_container = Frame(self._can_container)
        self._can_container.pack(side=LEFT, fill=BOTH, expand=1)
        x_scrollbar = Scrollbar(
            self_frm_scrollx, orient=HORIZONTAL, command=self._can_container.xview
        )
        x_scrollbar.pack(side=BOTTOM, fill=X)
        y_scrollbar = Scrollbar(
            self._frm_main, orient=VERTICAL, command=self._can_container.yview
        )
        y_scrollbar.pack(side=RIGHT, fill=Y)
        self._can_container.configure(xscrollcommand=x_scrollbar.set)
        self._can_container.configure(yscrollcommand=y_scrollbar.set)
        self._can_container.create_window((0, 0), window=self._frm_container, anchor=NW)
        self._can_container.bind(
            "<Configure>",
            lambda e: self._can_container.config(
                scrollregion=self._can_container.bbox(ALL)
            ),
        )

    def on_expand(self):
        """
        Automatically expand container canvas when sub-frames are resized.
        """

        self._can_container.event_generate("<Configure>")

    def _body(self):
        """
        Set up frame and widgets.
        """

        for i in range(4):
            self._frm_container.grid_columnconfigure(i, weight=1)
        self._frm_container.grid_rowconfigure(0, weight=1)

        self._frm_container.option_add("*Font", self.__app.font_sm)

        # serial port configuration panel
        self.frm_serial = SerialConfigFrame(
            self.__app,
            self._frm_container,
            GNSS,
            preselect=KNOWNGPS,
            timeouts=TIMEOUTS,
            bpsrates=BPSRATES,
            msgmodes=list(MSGMODES.keys()),
        )

        # socket client configuration panel
        self.frm_socketclient = SocketConfigFrame(
            self.__app,
            self._frm_container,
            GNSS,
        )

        # connection buttons
        self._frm_buttons = Frame(self._frm_container)
        self._btn_connect = Button(
            self._frm_buttons,
            width=45,
            height=35,
            image=self._img_serial,
            command=lambda: self._on_connect(CONNECTED),
        )
        self._lbl_connect = Label(self._frm_buttons, text="USB/UART")
        self._btn_connect_socket = Button(
            self._frm_buttons,
            width=45,
            height=35,
            image=self._img_socket,
            command=lambda: self._on_connect(CONNECTED_SOCKET),
        )
        self._lbl_connect_socket = Label(self._frm_buttons, text="TCP/UDP")
        self._btn_connect_file = Button(
            self._frm_buttons,
            width=45,
            height=35,
            image=self._img_dataread,
            command=lambda: self._on_connect(CONNECTED_FILE),
        )
        self._lbl_connect_file = Label(self._frm_buttons, text="FILE")
        self._btn_disconnect = Button(
            self._frm_buttons,
            width=45,
            height=35,
            image=self._img_disconn,
            command=lambda: self._on_connect(DISCONNECTED),
            state=DISABLED,
        )
        self._lbl_disconnect = Label(self._frm_buttons, text="STOP")
        self._btn_exit = Button(
            self._frm_buttons,
            width=45,
            height=35,
            image=self._img_exit,
            command=lambda: self.__app.on_exit(),
        )

        self._lbl_status_preset = Label(
            self._frm_buttons, font=self.__app.font_md2, text=""
        )

        # Other configuration options
        self._frm_options = Frame(self._frm_container)
        self._lbl_protocol = Label(self._frm_options, text=LBLPROTDISP)
        self._chk_nmea = Checkbutton(
            self._frm_options,
            text="NMEA",
            variable=self._prot_nmea,
        )
        self._chk_ubx = Checkbutton(
            self._frm_options,
            text="UBX",
            variable=self._prot_ubx,
        )
        self._chk_rtcm = Checkbutton(
            self._frm_options,
            text="RTCM",
            variable=self._prot_rtcm3,
        )
        self._chk_spartn = Checkbutton(
            self._frm_options,
            text="SPARTN",
            variable=self._prot_spartn,
        )
        self._chk_sbf = Checkbutton(
            self._frm_options,
            text="SBF",
            variable=self._prot_sbf,
        )
        self._chk_tty = Checkbutton(
            self._frm_options,
            text="TTY",
            variable=self._prot_tty,
        )
        self._lbl_consoledisplay = Label(self._frm_options, text=LBLDATADISP)
        self._spn_conformat = Spinbox(
            self._frm_options,
            values=FORMATS,
            width=10,
            state=READONLY,
            wrap=True,
            textvariable=self._console_format,
        )
        self._chk_tags = Checkbutton(
            self._frm_options,
            text="Tags",
            variable=self._colortag,
        )
        self._lbl_format = Label(self._frm_options, text=LBLDEGFORMAT)
        self._spn_format = Spinbox(
            self._frm_options,
            values=(DDD, DMS, DMM, ECEF),
            width=6,
            state=READONLY,
            wrap=True,
            textvariable=self._degrees_format,
        )
        self._spn_units = Spinbox(
            self._frm_options,
            values=(UMM, UIK, UI, UMK),
            width=13,
            state=READONLY,
            wrap=True,
            textvariable=self._units,
        )
        self._chk_scroll = Checkbutton(
            self._frm_options, text="Autoscroll", variable=self._autoscroll
        )
        self._spn_maxlines = Spinbox(
            self._frm_options,
            values=MAXLINES,
            width=6,
            wrap=True,
            textvariable=self._maxlines,
            state=READONLY,
        )
        self._chk_unusedsat = Checkbutton(
            self._frm_options, text=LBLSHOWUNUSED, variable=self._show_unusedsat
        )
        self._chk_datalog = Checkbutton(
            self._frm_options,
            text=LBLDATALOG,
            variable=self._datalog,
            command=lambda: self._on_data_log(),
        )
        self._spn_datalog = Spinbox(
            self._frm_options,
            values=(FORMATS),
            width=20,
            wrap=True,
            textvariable=self._logformat,
            state=READONLY,
        )
        self._chk_recordtrack = Checkbutton(
            self._frm_options,
            text=LBLTRACKRECORD,
            variable=self._record_track,
            command=lambda: self._on_record_track(),
        )
        self._chk_recorddatabase = Checkbutton(
            self._frm_options,
            text=LBLDATABASERECORD,
            variable=self._record_database,
            command=lambda: self._on_record_database(),
        )
        # configuration panel buttons
        self._lbl_ubxconfig = Label(
            self._frm_options,
            text=LBLUBXCONFIG,
        )
        self._btn_ubxconfig = Button(
            self._frm_options,
            width=45,
            image=self._img_ubxconfig,
            command=lambda: self._on_ubx_config(),
        )
        self._lbl_nmeaconfig = Label(
            self._frm_options,
            text=LBLNMEACONFIG,
        )
        self._btn_nmeaconfig = Button(
            self._frm_options,
            width=45,
            image=self._img_nmeaconfig,
            command=lambda: self._on_nmea_config(),
        )
        self._lbl_ntripconfig = Label(
            self._frm_options,
            text=LBLNTRIPCONFIG,
        )
        self._btn_ntripconfig = Button(
            self._frm_options,
            width=45,
            image=self._img_ntripconfig,
            command=lambda: self._on_ntrip_config(),
        )
        self._lbl_spartnconfig = Label(
            self._frm_options,
            text=LBLSPARTNCONFIG,
        )
        self._btn_spartnconfig = Button(
            self._frm_options,
            width=45,
            image=self._img_spartnconfig,
            command=lambda: self._on_spartn_config(),
        )
        # socket server configuration
        self.frm_socketserver = ServerConfigFrame(
            self.__app,
            self._frm_container,
        )

    def _do_layout(self):
        """
        Position widgets in frame.
        """

        self.frm_serial.grid(
            column=0, row=1, columnspan=4, padx=2, pady=2, sticky=(W, E)
        )
        ttk.Separator(self._frm_container).grid(
            column=0, row=2, columnspan=4, padx=2, pady=2, sticky=(W, E)
        )

        self.frm_socketclient.grid(
            column=0, row=3, columnspan=4, padx=2, pady=2, sticky=(W, E)
        )
        ttk.Separator(self._frm_container).grid(
            column=0, row=4, columnspan=4, padx=2, pady=2, sticky=(W, E)
        )

        self._frm_buttons.grid(column=0, row=5, columnspan=4, sticky=(W, E))
        self._btn_connect.grid(column=0, row=0, padx=2, pady=1)
        self._btn_connect_socket.grid(column=1, row=0, padx=2, pady=1)
        self._btn_connect_file.grid(column=2, row=0, padx=2, pady=1)
        self._btn_disconnect.grid(column=3, row=0, padx=2, pady=1)
        self._btn_exit.grid(column=4, row=0, padx=2, pady=1)
        self._lbl_connect.grid(column=0, row=1, padx=1, pady=1, sticky=(W, E))
        self._lbl_connect_socket.grid(column=1, row=1, padx=1, pady=1, sticky=(W, E))
        self._lbl_connect_file.grid(column=2, row=1, padx=1, pady=1, sticky=(W, E))
        self._lbl_disconnect.grid(column=3, row=1, padx=1, pady=1, sticky=(W, E))

        ttk.Separator(self._frm_container).grid(
            column=0, row=7, columnspan=4, padx=2, pady=2, sticky=(W, E)
        )

        self._frm_options.grid(column=0, row=8, columnspan=4, sticky=(W, E))
        self._lbl_protocol.grid(column=0, row=0, padx=2, pady=2, sticky=W)
        self._chk_nmea.grid(column=1, row=0, padx=0, pady=0, sticky=W)
        self._chk_ubx.grid(column=2, row=0, padx=0, pady=0, sticky=W)
        self._chk_rtcm.grid(column=1, row=1, padx=0, pady=0, sticky=W)
        self._chk_spartn.grid(column=2, row=1, padx=0, pady=0, sticky=W)
        self._chk_sbf.grid(column=3, row=0, padx=0, pady=0, sticky=W)
        self._chk_tty.grid(column=3, row=1, padx=0, pady=0, sticky=W)
        self._lbl_consoledisplay.grid(column=0, row=2, padx=2, pady=2, sticky=W)
        self._spn_conformat.grid(
            column=1, row=2, columnspan=2, padx=1, pady=2, sticky=W
        )
        self._chk_tags.grid(column=3, row=2, padx=1, pady=2, sticky=W)
        self._lbl_format.grid(column=0, row=3, padx=2, pady=2, sticky=W)
        self._spn_format.grid(column=1, row=3, padx=2, pady=2, sticky=W)
        self._spn_units.grid(column=2, row=3, columnspan=2, padx=2, pady=2, sticky=W)
        self._chk_scroll.grid(column=0, row=5, padx=2, pady=2, sticky=W)
        self._spn_maxlines.grid(column=1, row=5, columnspan=3, padx=2, pady=2, sticky=W)
        self._chk_unusedsat.grid(
            column=0, row=6, columnspan=2, padx=2, pady=2, sticky=W
        )
        self._chk_datalog.grid(column=0, row=7, padx=2, pady=2, sticky=W)
        self._spn_datalog.grid(column=1, row=7, columnspan=3, padx=2, pady=2, sticky=W)
        self._chk_recordtrack.grid(
            column=0, row=8, columnspan=2, padx=2, pady=2, sticky=W
        )
        self._chk_recorddatabase.grid(
            column=2, row=8, columnspan=2, padx=2, pady=2, sticky=W
        )
        self._btn_ubxconfig.grid(column=0, row=9)
        self._lbl_ubxconfig.grid(column=0, row=10)
        self._btn_nmeaconfig.grid(column=1, row=9)
        self._lbl_nmeaconfig.grid(column=1, row=10)
        self._btn_ntripconfig.grid(column=2, row=9)
        self._lbl_ntripconfig.grid(column=2, row=10)
        self._btn_spartnconfig.grid(column=3, row=9)
        self._lbl_spartnconfig.grid(column=3, row=10)
        ttk.Separator(self._frm_container).grid(
            column=0, row=10, columnspan=4, padx=2, pady=2, sticky=(W, E)
        )
        self.frm_socketserver.grid(
            column=0, row=11, columnspan=4, padx=2, pady=2, sticky=(W, E)
        )

    def reset(self):
        """
        Reset settings to saved configuration.
        """

        self._bind_events(False)
        cfg = self.__app.configuration
        self._prot_nmea.set(cfg.get("nmeaprot_b"))
        self._prot_ubx.set(cfg.get("ubxprot_b"))
        self._prot_sbf.set(cfg.get("sbfprot_b"))
        self._prot_rtcm3.set(cfg.get("rtcmprot_b"))
        self._prot_spartn.set(cfg.get("spartnprot_b"))
        self._prot_tty.set(cfg.get("ttyprot_b"))
        self._degrees_format.set(cfg.get("degreesformat_s"))
        self._colortag.set(cfg.get("colortag_b"))
        self._units.set(cfg.get("units_s"))
        self._autoscroll.set(cfg.get("autoscroll_b"))
        self._maxlines.set(cfg.get("maxlines_n"))
        self._console_format.set(cfg.get("consoleformat_s"))
        self.show_legend.set(cfg.get("legend_b"))
        self._show_unusedsat.set(cfg.get("unusedsat_b"))
        self._logformat.set(cfg.get("logformat_s"))
        self._datalog.set(cfg.get("datalog_b"))
        self.logpath = cfg.get("logpath_s")
        self._record_track.set(cfg.get("recordtrack_b"))
        self.trackpath = cfg.get("trackpath_s")
        self.databasepath = cfg.get("databasepath_s")
        if self.__app.db_enabled != SQLOK:
            self._record_database.set(0)
            self._chk_recorddatabase.config(state=DISABLED)
        else:
            self._record_database.set(cfg.get("database_b"))
        self.clients = 0
        self._bind_events(True)

    def _bind_events(self, add: bool = True):
        """
        Add or remove event bindings to/from widgets.

        :param bool add: add or remove binding
        """

        tracemode = "write"
        if add:
            self._prot_tty.trace_add(tracemode, self._on_update_tty)
        else:
            if len(self._prot_tty.trace_info()) > 0:
                self._prot_tty.trace_remove(
                    tracemode, self._prot_tty.trace_info()[0][1]
                )
        if add:
            self._prot_ubx.trace_add(tracemode, self._on_update_ubx)
        else:
            if len(self._prot_ubx.trace_info()) > 0:
                self._prot_ubx.trace_remove(
                    tracemode, self._prot_ubx.trace_info()[0][1]
                )
        if add:
            self._prot_sbf.trace_add(tracemode, self._on_update_sbf)
        else:
            if len(self._prot_sbf.trace_info()) > 0:
                self._prot_sbf.trace_remove(
                    tracemode, self._prot_sbf.trace_info()[0][1]
                )
        for setting in (
            self._prot_nmea,
            self._prot_rtcm3,
            self._prot_spartn,
            self._autoscroll,
            self._maxlines,
            self._units,
            self._degrees_format,
            self._console_format,
            self._datalog,
            self._logformat,
            self._record_track,
            self._record_database,
            self._show_unusedsat,
            self.show_legend,
            self._colortag,
        ):
            if add:
                setting.trace_add(tracemode, self._on_update_config)
            else:
                if len(setting.trace_info()) > 0:
                    setting.trace_remove(tracemode, setting.trace_info()[0][1])

    def _reset_frames(self):
        """
        Reset frames.
        """

        self.__app.frm_mapview.reset_map_refresh()
        self.__app.frm_spectrumview.reset()
        self.__app.reset_gnssstatus()

    def _on_update_ubx(self, var, index, mode):  # pylint: disable=unused-argument
        """
        UBX or SBF protocol mode has been updated.
        """

        try:
            self.update()
            if self._prot_ubx.get():
                self._prot_sbf.set(0)
                self.__app.configuration.set("sbfprot_b", int(self._prot_sbf.get()))
            self.__app.configuration.set("ubxprot_b", int(self._prot_ubx.get()))
            self._on_update_protocol()
        except (ValueError, TclError):
            pass

    def _on_update_sbf(self, var, index, mode):  # pylint: disable=unused-argument
        """
        UBX or SBF protocol mode has been updated.
        """

        try:
            self.update()
            if self._prot_sbf.get():
                self._prot_ubx.set(0)
                self.__app.configuration.set("ubxprot_b", int(self._prot_ubx.get()))
            self.__app.configuration.set("sbfprot_b", int(self._prot_sbf.get()))
            self._on_update_protocol()
        except (ValueError, TclError):
            pass

    def _on_update_tty(self, var, index, mode):  # pylint: disable=unused-argument
        """
        TTY mode has been updated.
        """

        try:
            tty = self._prot_tty.get()
            self.update()
            if tty:
                for wdg in (
                    self._prot_nmea,
                    self._prot_ubx,
                    self._prot_sbf,
                    self._prot_rtcm3,
                    self._prot_spartn,
                ):
                    wdg.set(0)
            else:
                self._prot_nmea.set(1)
                self._prot_ubx.set(1)
                self._prot_sbf.set(0)
                self._prot_rtcm3.set(1)
                self._prot_spartn.set(1)
            self.__app.configuration.set("ttyprot_b", tty)
        except (ValueError, TclError):
            pass

    def _on_update_config(self, var, index, mode):  # pylint: disable=unused-argument
        """
        Update in-memory configuration if setting is changed.
        """

        try:
            self.update()
            cfg = self.__app.configuration
            cfg.set("nmeaprot_b", int(self._prot_nmea.get()))
            cfg.set("rtcmprot_b", int(self._prot_rtcm3.get()))
            cfg.set("spartnprot_b", int(self._prot_spartn.get()))
            self._on_update_protocol()
            cfg.set("degreesformat_s", self._degrees_format.get())
            cfg.set("colortag_b", int(self._colortag.get()))
            cfg.set("units_s", self._units.get())
            cfg.set("autoscroll_b", int(self._autoscroll.get()))
            cfg.set("maxlines_n", int(self._maxlines.get()))
            cfg.set("consoleformat_s", self._console_format.get())
            cfg.set("legend_b", int(self.show_legend.get()))
            cfg.set("unusedsat_b", int(self._show_unusedsat.get()))
            cfg.set("datalog_b", int(self._datalog.get()))
            cfg.set("logformat_s", self._logformat.get())
            cfg.set("logpath_s", self.logpath)
            cfg.set("recordtrack_b", int(self._record_track.get()))
            cfg.set("trackpath_s", int(self.trackpath))
            cfg.set("database_b", int(self._record_database.get()))
            cfg.set("databasepath_s", self.databasepath)
        except (ValueError, TclError):
            pass

    def _on_update_protocol(self):
        """
        Protocol(s) have been updated.
        """

        self.__app.configuration.set(
            "protocol_n",
            NMEA_PROTOCOL * int(self._prot_nmea.get())
            + UBX_PROTOCOL * int(self._prot_ubx.get())
            + SBF_PROTOCOL * int(self._prot_sbf.get())
            + RTCM3_PROTOCOL * int(self._prot_rtcm3.get())
            + SPARTN_PROTOCOL * int(self._prot_spartn.get())
            + TTY_PROTOCOL * int(self._prot_tty.get()),
        )

    def _on_connect(self, conntype: int):
        """
        Start or stop connection (serial, socket or file).

        :param int conntype: connection type
        """

        connstr = ""
        conndict = {
            "protocol": self.__app.configuration.get("protocol_n"),
            "read_event": GNSS_EVENT,
            "eof_event": GNSS_EOF_EVENT,
            "timeout_event": GNSS_TIMEOUT_EVENT,
            "error_event": GNSS_ERR_EVENT,
            "inqueue": self.__app.gnss_inqueue,
            "outqueue": self.__app.gnss_outqueue,
            "socket_inqueue": self.__app.socket_inqueue,
            "conntype": conntype,
            "msgmode": self.frm_serial.msgmode,
            "inactivity_timeout": self.frm_serial.inactivity_timeout,
        }

        self.frm_socketserver.set_status(conntype)
        if conntype == CONNECTED:
            frm = self.frm_serial
            if frm.status == NOPORTS:
                return
            connstr = f"{frm.port}:{frm.port_desc} @ {frm.bpsrate}"
            conndict = dict(conndict, **{"serial_settings": frm})
            # poll for device software version on connection
            self.__app.poll_version(conndict["protocol"])
        elif conntype == CONNECTED_SOCKET:
            frm = self.frm_socketclient
            if not frm.valid_settings():
                self.__app.set_status("ERROR - invalid settings", ERRCOL)
                return
            connstr = f"{frm.server.get()}:{frm.port.get()}"
            conndict = dict(conndict, **{"socket_settings": frm})
        elif conntype == CONNECTED_FILE:
            self.infilepath = self.__app.file_handler.open_file(
                "datalog",
                (
                    ("datalog files", "*.log"),
                    ("u-center logs", "*.ubx"),
                    ("all files", "*.*"),
                ),
            )
            if self.infilepath is None:
                return
            connstr = f"{self.infilepath}"
            conndict = dict(conndict, **{"in_filepath": self.infilepath})
        elif conntype == DISCONNECTED:
            if self.__app.conn_status != DISCONNECTED:
                self.__app.conn_status = DISCONNECTED
                self.__app.stream_handler.stop_read_thread()
                return
        else:
            return

        self.__app.set_connection(connstr, OKCOL)
        self.__app.set_status("")
        self.__app.conn_status = conntype
        self._reset_frames()
        self.__app.stream_handler.start_read_thread(self.__app, conndict)

    def _on_ubx_config(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Open UBX configuration dialog panel.
        """

        self.__app.start_dialog(DLGTUBX)

    def _on_nmea_config(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Open NMEA configuration dialog panel.
        """

        self.__app.start_dialog(DLGTNMEA)

    def _on_ntrip_config(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Open NTRIP Client configuration dialog panel.
        """

        self.__app.start_dialog(DLGTNTRIP)

    def _on_spartn_config(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Open SPARTN Client configuration dialog panel.
        """

        self.__app.start_dialog(DLGTSPARTN)

    def _on_webmap(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Reset webmap refresh timer.
        """

        self.__app.frm_mapview.reset_map_refresh()

    def _on_data_log(self):
        """
        Start or stop data logger.
        """

        if self._datalog.get() == 1:
            if self.logpath in ("", None):
                self.logpath = self.__app.file_handler.set_logfile_path()
            if self.logpath is not None:
                self.__app.configuration.set("datalog_b", 1)
                self.__app.configuration.set("logpath_s", self.logpath)
                self.__app.set_status(f"Data logging enabled: {self.logpath}")
                if not self.__app.file_handler.open_logfile():
                    self.logpath = ""
                    self._datalog.set(0)
            else:
                self.logpath = ""
                self._datalog.set(0)
            self._spn_datalog.config(state=DISABLED)
        else:
            self.__app.configuration.set("datalog_b", 0)
            self._datalog.set(0)
            self.__app.file_handler.close_logfile()
            self.__app.set_status("Data logging disabled")
            self._spn_datalog.config(state=READONLY)

    def _on_record_track(self):
        """
        Start or stop track recorder.
        """

        if self._record_track.get() == 1:
            if self.trackpath in ("", None):
                self.trackpath = self.__app.file_handler.set_trackfile_path()
            if self.trackpath is not None:
                self.__app.configuration.set("recordtrack_b", 1)
                self.__app.configuration.set("trackpath_s", self.trackpath)
                self.__app.set_status(f"Track recording enabled: {self.trackpath}")
                if not self.__app.file_handler.open_trackfile():
                    self.trackpath = ""
                    self._record_track.set(0)
            else:
                self.trackpath = ""
                self._record_track.set(0)
        else:
            self._record_track.set(0)
            self.__app.configuration.set("recordtrack_b", 0)
            self.__app.file_handler.close_trackfile()
            self.__app.set_status("Track recording disabled")

    def _on_record_database(self):
        """
        Start or stop database recorder.
        """

        if self._record_database.get() == 1:
            if self.databasepath in ("", None):
                self.databasepath = self.__app.file_handler.set_database_path()
            if self.databasepath is not None:
                rc = self.__app.sqlite_handler.open(dbpath=self.databasepath)
                self.__app.configuration.set("database_b", rc == SQLOK)
                self.__app.configuration.set("databasepath_s", self.databasepath)
            else:
                self.databasepath = ""
                self._record_database.set(0)
        else:
            self.__app.configuration.set("database_b", 0)
            self.__app.set_status("Database recording disabled")

    def enable_controls(self, status: int):
        """
        Public method to enable or disable controls depending on
        connection status.

        :param int status: connection status as integer
               (0=Disconnected, 1=Connected to serial,
               2=Connected to file, 3=No serial ports available)

        """

        self.frm_serial.set_status(status)
        self.frm_socketclient.set_status(status)
        self.frm_socketserver.set_status(status)

        self._btn_connect.config(
            state=(
                DISABLED
                if status in (CONNECTED, CONNECTED_SOCKET, CONNECTED_FILE, NOPORTS)
                else NORMAL
            )
        )
        for ctl in (
            self._btn_connect_socket,
            self._btn_connect_file,
            self._chk_tty,
        ):
            ctl.config(
                state=(
                    DISABLED
                    if status in (CONNECTED, CONNECTED_SOCKET, CONNECTED_FILE)
                    else NORMAL
                )
            )
        self._btn_disconnect.config(
            state=(DISABLED if status in (DISCONNECTED,) else NORMAL)
        )

    def get_size(self) -> tuple:
        """
        Get current frame size.

        :return: (width, height)
        :rtype: tuple

        """

        self.update_idletasks()  # Make sure we know about any resizing
        return (self.winfo_width(), self.winfo_height())



"""
sbf_handler.py

SBF Protocol handler - handles all incoming SBF messages

Parses individual UBX messages (using pysbf2 library)
and adds selected attribute values to the app.gnss_status
data dictionary. This dictionary is then used to periodically
update the various user-selectable widget panels.

Created on 19 May 2025

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

import logging
from math import degrees

from pysbf2 import SBFMessage, itow2utc

from pygpsclient.globals import ASCII, BSR
from pygpsclient.helpers import fix2desc

DNUL = -2 * (10**10)
DNUS = 65535


class SBFHandler:
    """
    SBFHandler class
    """

    def __init__(self, app):
        """
        Constructor.

        :param Frame app: reference to main tkinter application
        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.appmaster  # Reference to root class (Tk)
        self.logger = logging.getLogger(__name__)

        self._cdb = 0
        self._raw_data = None
        self._parsed_data = None
        # Holds array of current satellites
        self.gsv_data = {}

    def process_data(self, raw_data: bytes, parsed_data: object):
        """
        Process relevant UBX message types

        :param bytes raw_data: raw data
        :param UBXMessage parsed_data: parsed data
        """

        if raw_data is None:
            return
        # self.logger.debug(f"data received {parsed_data.identity}")
        if parsed_data.identity == "PVTGeodetic":
            self._process_PVTGeodetic(parsed_data)
        elif parsed_data.identity == "ReceiverStatus":
            self._process_ReceiverStatus(parsed_data)
        elif parsed_data.identity == "ReceiverSetup":
            self._process_ReceiverSetup(parsed_data)

    def _process_PVTGeodetic(self, data: SBFMessage):
        """
        Process PVTGeodetic sentence - Position Velocity Track Geodetic.

        :param SBFMessage data: PVTGeodetic message
        """

        self.__app.gnss_status.utc = itow2utc(data.TOW)  # datetime.time
        if data.Latitude != DNUL:
            self.__app.gnss_status.lat = degrees(data.Latitude)
        if data.Longitude != DNUL:
            self.__app.gnss_status.lon = degrees(data.Longitude)
        if data.Height != DNUL:
            self.__app.gnss_status.hae = data.Height  # meters
        if data.HAccuracy != DNUS:
            self.__app.gnss_status.hacc = data.HAccuracy / 100  # meters
        if data.VAccuracy != DNUS:
            self.__app.gnss_status.vacc = data.VAccuracy / 100  # meters
        if data.NrSV != 255:
            self.__app.gnss_status.sip = data.NrSV
        if data.COG != DNUL:
            self.__app.gnss_status.track = data.COG
        self.__app.gnss_status.fix = fix2desc("PVTGeodetic", data.Type)
        if data.MeanCorrAge != 0:
            self.__app.gnss_status.diff_age = data.MeanCorrAge / 100  # seconds
        self.__app.gnss_status.diff_station = data.ReferenceID

    def _process_ReceiverStatus(self, data: SBFMessage):
        """
        Process ReceiverStatus sentence.

        :param SBFMessage data: ReceiverStatus message
        """

        sysdata = {}
        sysdata["cpuLoad"] = data.CPULoad
        sysdata["runTime"] = data.UpTime
        sysdata["tempValue"] = data.Temperature - 100  # offset by 100
        self.__app.gnss_status.sysmon_data = sysdata

    def _process_ReceiverSetup(self, data: SBFMessage):
        """
        Process ReceiverSetups sentence.

        :param SBFMessage data: ReceiverSetup message
        """

        verdata = {}
        verdata["hwversion"] = data.ProductName.decode(ASCII, errors=BSR).replace(
            "\x00", ""
        )
        verdata["fwversion"] = data.RxVersion.decode(ASCII, errors=BSR).replace(
            "\x00", ""
        )
        verdata["romversion"] = data.GNSSFWVersion.decode(ASCII, errors=BSR).replace(
            "\x00", ""
        )
        self.__app.gnss_status.version_data = verdata
